# SPDX-FileCopyrightText: Copyright (c) 2026 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
"""Typing for exotic float types (fp8, fp4, fp6, tf32).

Registers operator overloads for all SpecialFloatType instances except bf16
(which has its own typing via numba-cuda's bf16 registry).

Also registers FP8 constructors (fp8_e5m2, fp8_e4m3, fp8_e8m0),
cvt_e8m0_to_bf16raw, and bfloat16_raw attribute access.
"""

import operator
from numba_cuda_mlir.numba_cuda.typing.templates import (
    AttributeTemplate,
    ConcreteTemplate,
    Registry,
    signature,
)
from numba_cuda_mlir import types
from numba_cuda_mlir.numba_cuda.types.ext_types import bfloat16 as bf16

registry = Registry()
_fp8_globals_registered = False

_EXOTIC_FLOAT_TYPES = [
    types.f4E2M1FN,
    types.f6E2M3FN,
    types.f6E3M2FN,
    types.f8E3M4,
    types.f8E4M3B11FNUZ,
    types.f8E4M3FN,
    types.f8E4M3FNUZ,
    types.f8E4M3,
    types.f8E5M2FNUZ,
    types.f8E5M2,
    types.f8E8M0FNU,
    types.tf32,
]

_BINARY_OPS = [
    operator.add,
    operator.sub,
    operator.mul,
    operator.truediv,
    operator.iadd,
    operator.isub,
    operator.imul,
    operator.itruediv,
]

_COMPARISON_OPS = [
    operator.eq,
    operator.ne,
    operator.lt,
    operator.le,
    operator.gt,
    operator.ge,
]

_UNARY_OPS = [
    operator.neg,
    operator.pos,
]

_FP8_CTOR_INPUT_TYPES = [
    types.float16,
    bf16,
    types.float32,
    types.float64,
    types.int8,
    types.int16,
    types.int32,
    types.int64,
    types.uint8,
    types.uint16,
    types.uint32,
    types.uint64,
]

# Exotic float arithmetic/comparison/unary ops
for _ty in _EXOTIC_FLOAT_TYPES:
    for _op in _BINARY_OPS:

        class ExoticFloatBinaryTemplate(ConcreteTemplate):
            key = _op
            cases = [signature(_ty, _ty, _ty)]

        ExoticFloatBinaryTemplate.__name__ = f"ExoticFloat_{_ty.name}_{_op.__name__}"
        registry.register(ExoticFloatBinaryTemplate)
        registry.register_global(_op, types.Function(ExoticFloatBinaryTemplate))

    for _op in _COMPARISON_OPS:

        class ExoticFloatComparisonTemplate(ConcreteTemplate):
            key = _op
            cases = [signature(types.boolean, _ty, _ty)]

        ExoticFloatComparisonTemplate.__name__ = f"ExoticFloat_{_ty.name}_{_op.__name__}"
        registry.register(ExoticFloatComparisonTemplate)
        registry.register_global(_op, types.Function(ExoticFloatComparisonTemplate))

    for _op in _UNARY_OPS:

        class ExoticFloatUnaryTemplate(ConcreteTemplate):
            key = _op
            cases = [signature(_ty, _ty)]

        ExoticFloatUnaryTemplate.__name__ = f"ExoticFloat_{_ty.name}_{_op.__name__}"
        registry.register(ExoticFloatUnaryTemplate)
        registry.register_global(_op, types.Function(ExoticFloatUnaryTemplate))


def register_fp8_globals():
    """Register FP8 constructors, cvt_e8m0_to_bf16raw, and bfloat16_raw.x."""
    global _fp8_globals_registered
    if _fp8_globals_registered:
        return

    try:
        fp8_e5m2 = types.fp8_e5m2
        fp8_e4m3 = types.fp8_e4m3
        fp8_e8m0 = types.fp8_e8m0
        _type_fp8_e5m2 = types._type_fp8_e5m2
        _type_fp8_e4m3 = types._type_fp8_e4m3
        _type_fp8_e8m0 = types._type_fp8_e8m0
        cvt_e8m0_to_bf16raw = types.cvt_e8m0_to_bf16raw
        bfloat16_raw_type = types.bfloat16_raw_type
    except (AttributeError, ImportError):
        return

    for _py_type, _nb_type in [
        (fp8_e5m2, _type_fp8_e5m2),
        (fp8_e4m3, _type_fp8_e4m3),
        (fp8_e8m0, _type_fp8_e8m0),
    ]:

        class FP8CtorTemplate(ConcreteTemplate):
            key = _py_type
            cases = [signature(_nb_type)] + [
                signature(_nb_type, inp) for inp in _FP8_CTOR_INPUT_TYPES
            ]

        FP8CtorTemplate.__name__ = f"FP8Ctor_{_nb_type.name}"
        registry.register(FP8CtorTemplate)
        registry.register_global(_py_type, types.Function(FP8CtorTemplate))

    class CvtE8m0ToBf16rawTemplate(ConcreteTemplate):
        key = cvt_e8m0_to_bf16raw
        cases = [signature(bfloat16_raw_type, types.uint8)]

    registry.register(CvtE8m0ToBf16rawTemplate)
    registry.register_global(cvt_e8m0_to_bf16raw, types.Function(CvtE8m0ToBf16rawTemplate))

    class Bfloat16RawAttrTemplate(AttributeTemplate):
        key = bfloat16_raw_type

        def resolve_x(self, obj):
            return types.uint16

    registry.register_attr(Bfloat16RawAttrTemplate)
    _fp8_globals_registered = True



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause

import tokenize
import string


def parse_signature(sig):
    """Parse generalized ufunc signature.

    NOTE: ',' (COMMA) is a delimiter; not separator.
          This means trailing comma is legal.
    """

    def stripws(s):
        return "".join(c for c in s if c not in string.whitespace)

    def tokenizer(src):
        def readline():
            yield src

        gen = readline()
        return tokenize.generate_tokens(lambda: next(gen))

    def parse(src):
        tokgen = tokenizer(src)
        while True:
            tok = next(tokgen)
            if tok[1] == "(":
                symbols = []
                while True:
                    tok = next(tokgen)
                    if tok[1] == ")":
                        break
                    elif tok[0] == tokenize.NAME:
                        symbols.append(tok[1])
                    elif tok[1] == ",":
                        continue
                    else:
                        raise ValueError('bad token in signature "%s"' % tok[1])
                yield tuple(symbols)
                tok = next(tokgen)
                if tok[1] == ",":
                    continue
                elif tokenize.ISEOF(tok[0]):
                    break
            elif tokenize.ISEOF(tok[0]):
                break
            else:
                raise ValueError('bad token in signature "%s"' % tok[1])

    ins, _, outs = stripws(sig).partition("->")
    inputs = list(parse(ins))
    outputs = list(parse(outs))

    # check that all output symbols are defined in the inputs
    isym = set()
    osym = set()
    for grp in inputs:
        isym |= set(grp)
    for grp in outputs:
        osym |= set(grp)

    diff = osym.difference(isym)
    if diff:
        raise NameError("undefined output symbols: %s" % ",".join(sorted(diff)))

    return inputs, outputs



# SPDX-FileCopyrightText: Copyright (c) 2026 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
from numba_cuda_mlir.errors import ForceLiteralArg
import operator
import numpy as np
from numba_cuda_mlir.numba_cuda.typing.templates import (
    AttributeTemplate,
    AbstractTemplate,
    signature,
)
from numba_cuda_mlir import types
from numba_cuda_mlir.numba_cuda.typing import npydecl
from numba_cuda_mlir.numba_cuda.typing.npydecl import parse_dtype
from numba_cuda_mlir.lowering_utilities import type_conversions
from numba_cuda_mlir.numba_cuda.np.unsafe.ndarray import to_fixed_tuple

registry = npydecl.registry


def _dtype_arg(args, kws, pos, default):
    return args[pos] if len(args) > pos else kws.get("dtype", default)


@registry.register_global(to_fixed_tuple)
class ToFixedTupleTemplate(AbstractTemplate):
    """Typing template for to_fixed_tuple(array, length)"""

    def generic(self, args, kws):
        if len(args) != 2:
            return None

        array, length = args

        if not isinstance(array, types.Array) or array.ndim != 1:
            return None

        if not isinstance(length, types.IntegerLiteral):
            raise ForceLiteralArg({"length": types.IntegerLiteral})

        tuple_size = int(length.literal_value)
        tuple_type = types.UniTuple(dtype=array.dtype, count=tuple_size)
        return signature(tuple_type, array, length)


@registry.register_global(np.empty)
class NumpyEmptyTemplate(AbstractTemplate):
    """Typing template for np.empty"""

    def generic(self, args, kws):
        if len(args) < 1:
            return None

        shape = args[0]
        dtype = _dtype_arg(args, kws, 1, types.float64)

        # Determine dimensionality from shape
        if isinstance(shape, types.Integer):
            ndim = 1
        elif isinstance(shape, (types.UniTuple, types.Tuple)):
            ndim = shape.count if isinstance(shape, types.UniTuple) else len(shape.types)
        else:
            return None

        # Handle dtype
        if isinstance(dtype, types.DTypeSpec):
            element_type = dtype.dtype
        else:
            element_type = dtype

        restype = types.Array(dtype=element_type, ndim=ndim, layout="C")
        return signature(restype, *args)


@registry.register_global(np.zeros)
class NumpyZerosTemplate(AbstractTemplate):
    """Typing template for np.zeros"""

    def generic(self, args, kws):
        if len(args) < 1:
            return None

        shape = args[0]
        dtype = _dtype_arg(args, kws, 1, types.float64)

        if isinstance(shape, types.Integer):
            ndim = 1
        elif isinstance(shape, (types.UniTuple, types.Tuple)):
            ndim = shape.count if isinstance(shape, types.UniTuple) else len(shape.types)
        else:
            return None

        if isinstance(dtype, types.DTypeSpec):
            element_type = dtype.dtype
        else:
            element_type = dtype

        restype = types.Array(dtype=element_type, ndim=ndim, layout="C")
        return signature(restype, *args)


@registry.register_global(np.ones)
class NumpyOnesTemplate(AbstractTemplate):
    """Typing template for np.ones"""

    def generic(self, args, kws):
        if len(args) < 1:
            return None

        shape = args[0]
        dtype = _dtype_arg(args, kws, 1, types.float64)

        if isinstance(shape, types.Integer):
            ndim = 1
        elif isinstance(shape, (types.UniTuple, types.Tuple)):
            ndim = shape.count if isinstance(shape, types.UniTuple) else len(shape.types)
        else:
            return None

        if isinstance(dtype, types.DTypeSpec):
            element_type = dtype.dtype
        else:
            element_type = dtype

        restype = types.Array(dtype=element_type, ndim=ndim, layout="C")
        return signature(restype, *args)


@registry.register_global(np.full)
class NumpyFullTemplate(AbstractTemplate):
    """Typing template for np.full"""

    def generic(self, args, kws):
        if len(args) < 2:
            return None

        shape = args[0]
        value = args[1]
        dtype = _dtype_arg(args, kws, 2, types.float64)

        if isinstance(shape, types.Integer):
            ndim = 1
        elif isinstance(shape, (types.UniTuple, types.Tuple)):
            ndim = shape.count if isinstance(shape, types.UniTuple) else len(shape.types)
        else:
            return None

        if isinstance(dtype, types.DTypeSpec):
            element_type = dtype.dtype
        else:
            element_type = dtype

        restype = types.Array(dtype=element_type, ndim=ndim, layout="C")
        return signature(restype, *args)


# ============================================================================
# NumPy Binary Operations (np.add, np.subtract, np.multiply, np.divide)
# ============================================================================


@registry.register
class NumpyAddTemplate(AbstractTemplate):
    """Typing template for np.add"""

    key = np.add

    def generic(self, args, kws):
        if len(args) != 2:
            return None

        lhs, rhs = args

        # Handle scalar + scalar
        if isinstance(lhs, types.Number) and isinstance(rhs, types.Number):
            target_dtype = lhs.unify(self.context, rhs)
            if target_dtype is None:
                return None
            return signature(target_dtype, lhs, rhs)

        # Handle array operations
        if isinstance(lhs, types.Array) or isinstance(rhs, types.Array):
            lhs_ndim = lhs.ndim if isinstance(lhs, types.Array) else 0
            rhs_ndim = rhs.ndim if isinstance(rhs, types.Array) else 0
            target_ndim = max(lhs_ndim, rhs_ndim)

            if isinstance(lhs, types.Number):
                target_dtype = rhs.dtype.unify(self.context, lhs)
                target_layout = rhs.layout
            elif isinstance(rhs, types.Number):
                target_dtype = lhs.dtype.unify(self.context, rhs)
                target_layout = lhs.layout
            else:
                target_dtype = lhs.dtype.unify(self.context, rhs.dtype)
                target_layout = lhs.layout

            if target_dtype is None:
                return None

            restype = types.Array(dtype=target_dtype, ndim=target_ndim, layout=target_layout)
            return signature(restype, lhs, rhs)

        return None


@registry.register
class NumpySubtractTemplate(AbstractTemplate):
    """Typing template for np.subtract"""

    key = np.subtract

    def generic(self, args, kws):
        if len(args) != 2:
            return None

        lhs, rhs = args

        if isinstance(lhs, types.Number) and isinstance(rhs, types.Number):
            target_dtype = lhs.unify(self.context, rhs)
            if target_dtype is None:
                return None
            return signature(target_dtype, lhs, rhs)

        if isinstance(lhs, types.Array) or isinstance(rhs, types.Array):
            lhs_ndim = lhs.ndim if isinstance(lhs, types.Array) else 0
            rhs_ndim = rhs.ndim if isinstance(rhs, types.Array) else 0
            target_ndim = max(lhs_ndim, rhs_ndim)

            if isinstance(lhs, types.Number):
                target_dtype = rhs.dtype.unify(self.context, lhs)
                target_layout = rhs.layout
            elif isinstance(rhs, types.Number):
                target_dtype = lhs.dtype.unify(self.context, rhs)
                target_layout = lhs.layout
            else:
                target_dtype = lhs.dtype.unify(self.context, rhs.dtype)
                target_layout = lhs.layout

            if target_dtype is None:
                return None

            restype = types.Array(dtype=target_dtype, ndim=target_ndim, layout=target_layout)
            return signature(restype, lhs, rhs)

        return None


@registry.register
class NumpyMultiplyTemplate(AbstractTemplate):
    """Typing template for np.multiply"""

    key = np.multiply

    def generic(self, args, kws):
        if len(args) != 2:
            return None

        lhs, rhs = args

        if isinstance(lhs, types.Number) and isinstance(rhs, types.Number):
            target_dtype = lhs.unify(self.context, rhs)
            if target_dtype is None:
                return None
            return signature(target_dtype, lhs, rhs)

        if isinstance(lhs, types.Array) or isinstance(rhs, types.Array):
            lhs_ndim = lhs.ndim if isinstance(lhs, types.Array) else 0
            rhs_ndim = rhs.ndim if isinstance(rhs, types.Array) else 0
            target_ndim = max(lhs_ndim, rhs_ndim)

            if isinstance(lhs, types.Number):
                target_dtype = rhs.dtype.unify(self.context, lhs)
                target_layout = rhs.layout
            elif isinstance(rhs, types.Number):
                target_dtype = lhs.dtype.unify(self.context, rhs)
                target_layout = lhs.layout
            else:
                target_dtype = lhs.dtype.unify(self.context, rhs.dtype)
                target_layout = lhs.layout

            if target_dtype is None:
                return None

            restype = types.Array(dtype=target_dtype, ndim=target_ndim, layout=target_layout)
            return signature(restype, lhs, rhs)

        return None


@registry.register
class NumpyDivideTemplate(AbstractTemplate):
    """Typing template for np.divide"""

    key = np.divide

    def generic(self, args, kws):
        if len(args) != 2:
            return None

        lhs, rhs = args

        if isinstance(lhs, types.Number) and isinstance(rhs, types.Number):
            target_dtype = lhs.unify(self.context, rhs)
            if target_dtype is None:
                return None
            return signature(target_dtype, lhs, rhs)

        if isinstance(lhs, types.Array) or isinstance(rhs, types.Array):
            lhs_ndim = lhs.ndim if isinstance(lhs, types.Array) else 0
            rhs_ndim = rhs.ndim if isinstance(rhs, types.Array) else 0
            target_ndim = max(lhs_ndim, rhs_ndim)

            if isinstance(lhs, types.Number):
                target_dtype = rhs.dtype.unify(self.context, lhs)
                target_layout = rhs.layout
            elif isinstance(rhs, types.Number):
                target_dtype = lhs.dtype.unify(self.context, rhs)
                target_layout = lhs.layout
            else:
                target_dtype = lhs.dtype.unify(self.context, rhs.dtype)
                target_layout = lhs.layout

            if target_dtype is None:
                return None

            restype = types.Array(dtype=target_dtype, ndim=target_ndim, layout=target_layout)
            return signature(restype, lhs, rhs)

        return None


@registry.register
class NumpyNegativeTemplate(AbstractTemplate):
    """Typing template for np.negative"""

    key = np.negative

    def generic(self, args, kws):
        if len(args) != 1:
            return None

        x = args[0]

        if isinstance(x, types.Array):
            restype = types.Array(dtype=x.dtype, ndim=x.ndim, layout=x.layout)
            return signature(restype, x)

        return None


# ============================================================================
# NumPy Unary Math Functions (absolute, ceil, floor, log, exp)
# ============================================================================


@registry.register
class NumpyAbsoluteTemplate(AbstractTemplate):
    """Typing template for np.absolute"""

    key = np.absolute

    def generic(self, args, kws):
        if len(args) != 1:
            return None

        x = args[0]

        if isinstance(x, types.Number):
            return signature(x, x)
        elif isinstance(x, types.Array):
            restype = types.Array(dtype=x.dtype, ndim=x.ndim, layout=x.layout)
            return signature(restype, x)

        return None


@registry.register
class NumpyCeilTemplate(AbstractTemplate):
    """Typing template for np.ceil"""

    key = np.ceil

    def generic(self, args, kws):
        if len(args) != 1:
            return None

        x = args[0]

        if isinstance(x, types.Number):
            return signature(x, x)
        elif isinstance(x, types.Array):
            restype = types.Array(dtype=x.dtype, ndim=x.ndim, layout=x.layout)
            return signature(restype, x)

        return None


@registry.register
class NumpyFloorTemplate(AbstractTemplate):
    """Typing template for np.floor"""

    key = np.floor

    def generic(self, args, kws):
        if len(args) != 1:
            return None

        x = args[0]

        if isinstance(x, types.Number):
            return signature(x, x)
        elif isinstance(x, types.Array):
            restype = types.Array(dtype=x.dtype, ndim=x.ndim, layout=x.layout)
            return signature(restype, x)

        return None


@registry.register
class NumpyLogTemplate(AbstractTemplate):
    """Typing template for np.log"""

    key = np.log

    def generic(self, args, kws):
        if len(args) != 1:
            return None

        x = args[0]

        if isinstance(x, types.Number):
            return signature(types.float64, x)
        elif isinstance(x, types.Array):
            restype = types.Array(dtype=types.float64, ndim=x.ndim, layout=x.layout)
            return signature(restype, x)

        return None


@registry.register
class NumpyExpTemplate(AbstractTemplate):
    """Typing template for np.exp"""

    key = np.exp

    def generic(self, args, kws):
        if len(args) != 1:
            return None

        x = args[0]

        if isinstance(x, types.Number):
            return signature(types.float64, x)
        elif isinstance(x, types.Array):
            restype = types.Array(dtype=types.float64, ndim=x.ndim, layout=x.layout)
            return signature(restype, x)

        return None


# ============================================================================
# NumPy Matrix Operations (matmul, dot, transpose)
# ============================================================================


@registry.register
class NumpyMatmulTemplate(AbstractTemplate):
    """Typing template for np.matmul"""

    key = np.matmul

    def generic(self, args, kws):
        if len(args) != 2:
            return None

        lhs, rhs = args

        if isinstance(lhs, types.Array) and isinstance(rhs, types.Array):
            if lhs.ndim == 2 and rhs.ndim == 2:
                target_dtype = lhs.dtype.unify(self.context, rhs.dtype)
                if target_dtype is None:
                    return None

                restype = types.Array(dtype=target_dtype, ndim=2, layout=lhs.layout)
                return signature(restype, lhs, rhs)

        return None


@registry.register_global(np.dot)
class NumpyDotTemplate(AbstractTemplate):
    """Typing template for np.dot"""

    def generic(self, args, kws):
        if len(args) != 2:
            return None

        lhs, rhs = args

        # Scalar * scalar
        if isinstance(lhs, types.Number) and isinstance(rhs, types.Number):
            target_dtype = lhs.unify(self.context, rhs)
            if target_dtype is None:
                return None
            return signature(target_dtype, lhs, rhs)

        # Scalar * array or array * scalar
        if (isinstance(lhs, types.Number) and isinstance(rhs, types.Array)) or (
            isinstance(rhs, types.Number) and isinstance(lhs, types.Array)
        ):
            if isinstance(lhs, types.Number):
                array_val = rhs
                scalar_val = lhs
            else:
                array_val = lhs
                scalar_val = rhs

            target_dtype = array_val.dtype.unify(self.context, scalar_val)
            if target_dtype is None:
                return None

            restype = types.Array(dtype=target_dtype, ndim=array_val.ndim, layout=array_val.layout)
            return signature(restype, lhs, rhs)

        # Array operations
        if isinstance(lhs, types.Array) and isinstance(rhs, types.Array):
            # 1D @ 1D -> scalar
            if lhs.ndim == 1 and rhs.ndim == 1:
                target_dtype = lhs.dtype.unify(self.context, rhs.dtype)
                if target_dtype is None:
                    return None
                return signature(target_dtype, lhs, rhs)

            # 2D @ 2D -> 2D
            if lhs.ndim == 2 and rhs.ndim == 2:
                target_dtype = lhs.dtype.unify(self.context, rhs.dtype)
                if target_dtype is None:
                    return None
                restype = types.Array(dtype=target_dtype, ndim=2, layout=lhs.layout)
                return signature(restype, lhs, rhs)

            # ND @ 1D or 1D @ ND or ND @ MD
            if lhs.ndim >= 2 and rhs.ndim == 1:
                restype = types.Array(dtype=lhs.dtype, ndim=lhs.ndim - 1, layout=lhs.layout)
                return signature(restype, lhs, rhs)

            if lhs.ndim >= 1 and rhs.ndim >= 2:
                restype = types.Array(
                    dtype=lhs.dtype, ndim=lhs.ndim + rhs.ndim - 2, layout=lhs.layout
                )
                return signature(restype, lhs, rhs)

        return None


@registry.register_global(np.transpose)
class NumpyTransposeTemplate(AbstractTemplate):
    """Typing template for np.transpose"""

    def generic(self, args, kws):
        if len(args) != 1:
            return None

        x = args[0]

        if isinstance(x, types.Array):
            restype = types.Array(dtype=x.dtype, ndim=x.ndim, layout=x.layout)
            return signature(restype, x)

        return None


# ============================================================================
# NumPy Reduction and Statistical Functions
# ============================================================================


@registry.register_global(np.sum)
class NumpySumTemplate(AbstractTemplate):
    """Typing template for np.sum"""

    def generic(self, args, kws):
        if len(args) != 1:
            return None

        arr = args[0]

        if isinstance(arr, types.Array):
            # Sum returns scalar of same dtype
            return signature(arr.dtype, arr)

        return None


@registry.register_global(np.any)
class NumpyAnyTemplate(AbstractTemplate):
    """Typing template for np.any"""

    def generic(self, args, kws):
        if len(args) != 1:
            return None

        arr = args[0]

        if isinstance(arr, types.Array):
            return signature(types.boolean, arr)

        return None


@registry.register_global(np.all)
class NumpyAllTemplate(AbstractTemplate):
    """Typing template for np.all"""

    def generic(self, args, kws):
        if len(args) != 1:
            return None

        arr = args[0]

        if isinstance(arr, types.Array):
            return signature(types.boolean, arr)

        return None


@registry.register_global(np.var)
class NumpyVarTemplate(AbstractTemplate):
    """Typing template for np.var"""

    def generic(self, args, kws):
        if len(args) != 1:
            return None

        arr = args[0]

        if isinstance(arr, types.Array):
            return signature(arr.dtype, arr)

        return None


@registry.register_global(np.mean)
class NumpyMeanTemplate(AbstractTemplate):
    """Typing template for np.mean"""

    def generic(self, args, kws):
        if len(args) != 1:
            return None

        arr = args[0]

        if isinstance(arr, types.Array):
            return signature(arr.dtype, arr)

        return None


@registry.register_global(np.min)
class NumpyMinTemplate(AbstractTemplate):
    """Typing template for np.min"""

    def generic(self, args, kws):
        if len(args) != 1:
            return None

        arr = args[0]

        if isinstance(arr, types.Array):
            return signature(arr.dtype, arr)

        return None


@registry.register_global(np.max)
class NumpyMaxTemplate(AbstractTemplate):
    """Typing template for np.max"""

    def generic(self, args, kws):
        if len(args) != 1:
            return None

        arr = args[0]

        if isinstance(arr, types.Array):
            return signature(arr.dtype, arr)

        return None


@registry.register_global(np.prod)
class NumpyProdTemplate(AbstractTemplate):
    """Typing template for np.prod"""

    def generic(self, args, kws):
        if len(args) != 1:
            return None

        arr = args[0]

        if isinstance(arr, types.Array):
            return signature(arr.dtype, arr)

        return None


@registry.register_global(np.nanmin)
class NumpyNanMinTemplate(AbstractTemplate):
    """Typing template for np.nanmin"""

    def generic(self, args, kws):
        if len(args) != 1:
            return None
        arr = args[0]
        if isinstance(arr, types.Array):
            return signature(arr.dtype, arr)
        return None


@registry.register_global(np.nanmax)
class NumpyNanMaxTemplate(AbstractTemplate):
    """Typing template for np.nanmax"""

    def generic(self, args, kws):
        if len(args) != 1:
            return None
        arr = args[0]
        if isinstance(arr, types.Array):
            return signature(arr.dtype, arr)
        return None


@registry.register_global(np.nansum)
class NumpyNanSumTemplate(AbstractTemplate):
    """Typing template for np.nansum"""

    def generic(self, args, kws):
        if len(args) != 1:
            return None
        arr = args[0]
        if isinstance(arr, types.Array):
            return signature(arr.dtype, arr)
        return None


@registry.register_global(np.nanprod)
class NumpyNanProdTemplate(AbstractTemplate):
    """Typing template for np.nanprod"""

    def generic(self, args, kws):
        if len(args) != 1:
            return None
        arr = args[0]
        if isinstance(arr, types.Array):
            return signature(arr.dtype, arr)
        return None


@registry.register_global(np.nanmean)
class NumpyNanMeanTemplate(AbstractTemplate):
    """Typing template for np.nanmean"""

    def generic(self, args, kws):
        if len(args) != 1:
            return None
        arr = args[0]
        if isinstance(arr, types.Array):
            return signature(arr.dtype, arr)
        return None


@registry.register
class NumpySqrtTemplate(AbstractTemplate):
    """Typing template for np.sqrt"""

    key = np.sqrt

    def generic(self, args, kws):
        if len(args) != 1:
            return None

        x = args[0]

        if isinstance(x, types.Array):
            restype = types.Array(dtype=x.dtype, ndim=x.ndim, layout=x.layout)
            return signature(restype, x)

        return None


@registry.register
class NumpyAbsTemplate(AbstractTemplate):
    """Typing template for np.abs"""

    key = np.abs

    def generic(self, args, kws):
        if len(args) != 1:
            return None

        x = args[0]

        if isinstance(x, types.Array):
            restype = types.Array(dtype=x.dtype, ndim=x.ndim, layout=x.layout)
            return signature(restype, x)

        return None


@registry.register_global(np.reshape)
class NumpyReshapeTemplate(AbstractTemplate):
    """Typing template for np.reshape"""

    def generic(self, args, kws):
        if len(args) != 2:
            return None

        arr, shape = args

        if isinstance(arr, types.Array) and isinstance(shape, (types.Tuple, types.Array)):
            # Reshape returns same dtype, potentially different ndim
            # For now, return same array type
            return signature(arr, arr, shape)

        return None


@registry.register_global(len)
class LenTemplate(AbstractTemplate):
    """Typing template for len() with arrays and tuples"""

    def generic(self, args, kws):
        if len(args) != 1:
            return None

        arg = args[0]

        if isinstance(arg, (types.Array, types.Tuple, types.UniTuple)):
            return signature(types.intp, arg)

        return None


# ============================================================================
# NumPy Array Methods
# ============================================================================


@registry.register_attr
class ArrayAttributeTemplate(AttributeTemplate):
    """Typing template for array methods"""

    key = types.Array

    def resolve_size(self, arr):
        """Type arr.size -> int (property, not method)"""
        return types.intp

    def resolve_sum(self, arr):
        """Type arr.sum() -> scalar of same dtype"""
        return types.BoundFunction(NumpyArraySumMethodTemplate, arr)

    def resolve_prod(self, arr):
        """Type arr.prod() -> scalar of same dtype"""
        return types.BoundFunction(NumpyArrayProdMethodTemplate, arr)

    def resolve_min(self, arr):
        """Type arr.min() -> scalar of same dtype"""
        return types.BoundFunction(NumpyArrayMinMethodTemplate, arr)

    def resolve_max(self, arr):
        """Type arr.max() -> scalar of same dtype"""
        return types.BoundFunction(NumpyArrayMaxMethodTemplate, arr)

    def resolve_reshape(self, arr):
        """Type arr.reshape(shape) -> array"""
        return types.BoundFunction(NumpyArrayReshapeMethodTemplate, arr)

    def resolve_view(self, arr):
        """Type arr.view(dtype) -> array"""
        return types.BoundFunction(NumpyArrayViewMethodTemplate, arr)

    def resolve_ctypes(self, arr):
        """Type arr.ctypes -> ArrayCTypes"""
        return types.ArrayCTypes(arr)

    def generic_resolve(self, arr, attr):
        """Handle record array field access: arr.field_name -> array view"""
        if isinstance(arr.dtype, types.Record):
            record_type = arr.dtype
            if attr in record_type.fields:
                field_info = record_type.fields[attr]
                field_dtype = field_info.type
                return types.Array(
                    dtype=field_dtype,
                    ndim=arr.ndim,
                    layout="A",
                )
        return None


@registry.register_attr
class ArrayCTypesAttributeTemplate(AttributeTemplate):
    """Typing template for ArrayCTypes attributes"""

    key = types.ArrayCTypes

    def resolve_data(self, ctinfo):
        """Type ctinfo.data -> uintp (pointer as integer)"""
        return types.uintp


class NumpyArraySumMethodTemplate(AbstractTemplate):
    """Typing for arr.sum() method"""

    key = "Array.sum"

    def generic(self, args, kws):
        # arr.sum() takes no arguments
        if len(args) == 0:
            # Get the this (array) from the bound function
            arr = self.this
            return signature(arr.dtype, recvr=self.this)
        return None


class NumpyArrayProdMethodTemplate(AbstractTemplate):
    """Typing for arr.prod() method"""

    key = "Array.prod"

    def generic(self, args, kws):
        if len(args) == 0:
            arr = self.this
            return signature(arr.dtype, recvr=self.this)
        return None


class NumpyArrayMinMethodTemplate(AbstractTemplate):
    """Typing for arr.min() method"""

    key = "Array.min"

    def generic(self, args, kws):
        if len(args) == 0:
            arr = self.this
            return signature(arr.dtype, recvr=self.this)
        return None


class NumpyArrayMaxMethodTemplate(AbstractTemplate):
    """Typing for arr.max() method"""

    key = "Array.max"

    def generic(self, args, kws):
        if len(args) == 0:
            arr = self.this
            return signature(arr.dtype, recvr=self.this)
        return None


class NumpyArrayReshapeMethodTemplate(AbstractTemplate):
    """Typing for arr.reshape(shape) method"""

    key = "Array.reshape"

    def generic(self, args, kws):
        if len(args) == 1:
            arr = self.this
            shape = args[0]
            # Return same array type for now
            return signature(arr, shape, recvr=self.this)
        return None


class NumpyArrayViewMethodTemplate(AbstractTemplate):
    """Typing for arr.view(dtype) method"""

    key = "Array.view"

    def generic(self, args, kws):
        if len(args) == 1:
            arr = self.this
            dtype = args[0]
            element_type = dtype.dtype if isinstance(dtype, types.DTypeSpec) else dtype
            ret = types.Array(element_type, arr.ndim, arr.layout)
            return signature(ret, dtype, recvr=self.this)
        return None


# ============================================================================
# Python Operator Overloads for Arrays
# ============================================================================


@registry.register_global(operator.add)
class OperatorAddTemplate(AbstractTemplate):
    """Typing template for operator.add with arrays"""

    def generic(self, args, kws):
        if len(args) != 2:
            return None

        lhs, rhs = args

        # If one of the operands is an array, handle array addition
        if isinstance(lhs, types.Array) or isinstance(rhs, types.Array):
            lhs_ndim = lhs.ndim if isinstance(lhs, types.Array) else 0
            rhs_ndim = rhs.ndim if isinstance(rhs, types.Array) else 0
            target_ndim = max(lhs_ndim, rhs_ndim)

            # Determine target dtype and layout
            if isinstance(lhs, types.Number):
                target_dtype = rhs.dtype.unify(self.context, lhs)
                target_layout = rhs.layout
            elif isinstance(rhs, types.Number):
                target_dtype = lhs.dtype.unify(self.context, rhs)
                target_layout = lhs.layout
            else:
                target_dtype = lhs.dtype.unify(self.context, rhs.dtype)
                target_layout = lhs.layout

            if target_dtype is None:
                return None

            restype = types.Array(dtype=target_dtype, ndim=target_ndim, layout=target_layout)
            return signature(restype, lhs, rhs)

        return None


@registry.register_global(operator.iadd)
class OperatorIAddTemplate(AbstractTemplate):
    """Typing template for operator.iadd with arrays"""

    def generic(self, args, kws):
        if len(args) != 2:
            return None

        lhs, rhs = args

        if isinstance(lhs, types.Array) or isinstance(rhs, types.Array):
            lhs_ndim = lhs.ndim if isinstance(lhs, types.Array) else 0
            rhs_ndim = rhs.ndim if isinstance(rhs, types.Array) else 0
            target_ndim = max(lhs_ndim, rhs_ndim)

            if isinstance(lhs, types.Number):
                target_dtype = rhs.dtype.unify(self.context, lhs)
                target_layout = rhs.layout
            elif isinstance(rhs, types.Number):
                target_dtype = lhs.dtype.unify(self.context, rhs)
                target_layout = lhs.layout
            else:
                target_dtype = lhs.dtype.unify(self.context, rhs.dtype)
                target_layout = lhs.layout

            if target_dtype is None:
                return None

            restype = types.Array(dtype=target_dtype, ndim=target_ndim, layout=target_layout)
            return signature(restype, lhs, rhs)

        return None


@registry.register_global(operator.sub)
class OperatorSubTemplate(AbstractTemplate):
    """Typing template for operator.sub with arrays"""

    def generic(self, args, kws):
        if len(args) != 2:
            return None

        lhs, rhs = args

        if isinstance(lhs, types.Array) or isinstance(rhs, types.Array):
            lhs_ndim = lhs.ndim if isinstance(lhs, types.Array) else 0
            rhs_ndim = rhs.ndim if isinstance(rhs, types.Array) else 0
            target_ndim = max(lhs_ndim, rhs_ndim)

            if isinstance(lhs, types.Number):
                target_dtype = rhs.dtype.unify(self.context, lhs)
                target_layout = rhs.layout
            elif isinstance(rhs, types.Number):
                target_dtype = lhs.dtype.unify(self.context, rhs)
                target_layout = lhs.layout
            else:
                target_dtype = lhs.dtype.unify(self.context, rhs.dtype)
                target_layout = lhs.layout

            if target_dtype is None:
                return None

            restype = types.Array(dtype=target_dtype, ndim=target_ndim, layout=target_layout)
            return signature(restype, lhs, rhs)

        return None


@registry.register_global(operator.isub)
class OperatorISubTemplate(AbstractTemplate):
    """Typing template for operator.isub with arrays"""

    def generic(self, args, kws):
        if len(args) != 2:
            return None

        lhs, rhs = args

        if isinstance(lhs, types.Array) or isinstance(rhs, types.Array):
            lhs_ndim = lhs.ndim if isinstance(lhs, types.Array) else 0
            rhs_ndim = rhs.ndim if isinstance(rhs, types.Array) else 0
            target_ndim = max(lhs_ndim, rhs_ndim)

            if isinstance(lhs, types.Number):
                target_dtype = rhs.dtype.unify(self.context, lhs)
                target_layout = rhs.layout
            elif isinstance(rhs, types.Number):
                target_dtype = lhs.dtype.unify(self.context, rhs)
                target_layout = lhs.layout
            else:
                target_dtype = lhs.dtype.unify(self.context, rhs.dtype)
                target_layout = lhs.layout

            if target_dtype is None:
                return None

            restype = types.Array(dtype=target_dtype, ndim=target_ndim, layout=target_layout)
            return signature(restype, lhs, rhs)

        return None


@registry.register_global(operator.mul)
class OperatorMulTemplate(AbstractTemplate):
    """Typing template for operator.mul with arrays"""

    def generic(self, args, kws):
        if len(args) != 2:
            return None

        lhs, rhs = args

        if isinstance(lhs, types.Array) or isinstance(rhs, types.Array):
            lhs_ndim = lhs.ndim if isinstance(lhs, types.Array) else 0
            rhs_ndim = rhs.ndim if isinstance(rhs, types.Array) else 0
            target_ndim = max(lhs_ndim, rhs_ndim)

            if isinstance(lhs, types.Number):
                target_dtype = rhs.dtype.unify(self.context, lhs)
                target_layout = rhs.layout
            elif isinstance(rhs, types.Number):
                target_dtype = lhs.dtype.unify(self.context, rhs)
                target_layout = lhs.layout
            else:
                target_dtype = lhs.dtype.unify(self.context, rhs.dtype)
                target_layout = lhs.layout

            if target_dtype is None:
                return None

            restype = types.Array(dtype=target_dtype, ndim=target_ndim, layout=target_layout)
            return signature(restype, lhs, rhs)

        return None


@registry.register_global(operator.imul)
class OperatorIMulTemplate(AbstractTemplate):
    """Typing template for operator.imul with arrays"""

    def generic(self, args, kws):
        if len(args) != 2:
            return None

        lhs, rhs = args

        if isinstance(lhs, types.Array) or isinstance(rhs, types.Array):
            lhs_ndim = lhs.ndim if isinstance(lhs, types.Array) else 0
            rhs_ndim = rhs.ndim if isinstance(rhs, types.Array) else 0
            target_ndim = max(lhs_ndim, rhs_ndim)

            if isinstance(lhs, types.Number):
                target_dtype = rhs.dtype.unify(self.context, lhs)
                target_layout = rhs.layout
            elif isinstance(rhs, types.Number):
                target_dtype = lhs.dtype.unify(self.context, rhs)
                target_layout = lhs.layout
            else:
                target_dtype = lhs.dtype.unify(self.context, rhs.dtype)
                target_layout = lhs.layout

            if target_dtype is None:
                return None

            restype = types.Array(dtype=target_dtype, ndim=target_ndim, layout=target_layout)
            return signature(restype, lhs, rhs)

        return None


@registry.register_global(operator.truediv)
class OperatorTruedivTemplate(AbstractTemplate):
    """Typing template for operator.truediv with arrays"""

    def generic(self, args, kws):
        if len(args) != 2:
            return None

        lhs, rhs = args

        if isinstance(lhs, types.Array) or isinstance(rhs, types.Array):
            lhs_ndim = lhs.ndim if isinstance(lhs, types.Array) else 0
            rhs_ndim = rhs.ndim if isinstance(rhs, types.Array) else 0
            target_ndim = max(lhs_ndim, rhs_ndim)

            # Get dtypes
            lhs_dtype = lhs.dtype if isinstance(lhs, types.Array) else lhs
            rhs_dtype = rhs.dtype if isinstance(rhs, types.Array) else rhs

            # True division always returns float
            # Get the unified type first
            target_dtype = lhs_dtype.unify(self.context, rhs_dtype)
            if target_dtype is None:
                return None

            # Convert to float if integer
            if isinstance(target_dtype, types.Integer):
                # Use at least float32 for integer division
                bitwidth = max(target_dtype.bitwidth, 32)
                target_dtype = type_conversions.float_of_width(bitwidth)

            # Get layout
            if isinstance(lhs, types.Array):
                target_layout = lhs.layout
            elif isinstance(rhs, types.Array):
                target_layout = rhs.layout
            else:
                target_layout = "C"

            restype = types.Array(dtype=target_dtype, ndim=target_ndim, layout=target_layout)
            return signature(restype, lhs, rhs)

        return None


@registry.register_global(operator.itruediv)
class OperatorITruedivTemplate(AbstractTemplate):
    """Typing template for operator.itruediv with arrays"""

    def generic(self, args, kws):
        if len(args) != 2:
            return None

        lhs, rhs = args

        if isinstance(lhs, types.Array) or isinstance(rhs, types.Array):
            lhs_ndim = lhs.ndim if isinstance(lhs, types.Array) else 0
            rhs_ndim = rhs.ndim if isinstance(rhs, types.Array) else 0
            target_ndim = max(lhs_ndim, rhs_ndim)

            if isinstance(lhs, types.Number):
                target_dtype = rhs.dtype.unify(self.context, lhs)
                target_layout = rhs.layout
            elif isinstance(rhs, types.Number):
                target_dtype = lhs.dtype.unify(self.context, rhs)
                target_layout = lhs.layout
            else:
                target_dtype = lhs.dtype.unify(self.context, rhs.dtype)
                target_layout = lhs.layout

            if target_dtype is None:
                return None

            restype = types.Array(dtype=target_dtype, ndim=target_ndim, layout=target_layout)
            return signature(restype, lhs, rhs)

        return None


@registry.register_global(operator.neg)
class OperatorNegTemplate(AbstractTemplate):
    """Typing template for operator.neg with arrays"""

    def generic(self, args, kws):
        if len(args) != 1:
            return None

        x = args[0]

        if isinstance(x, types.Array):
            restype = types.Array(dtype=x.dtype, ndim=x.ndim, layout=x.layout)
            return signature(restype, x)

        return None


@registry.register_global(abs)
class AbsTemplate(AbstractTemplate):
    """Typing template for abs() with arrays"""

    def generic(self, args, kws):
        if len(args) != 1:
            return None

        x = args[0]

        if isinstance(x, types.Array):
            restype = types.Array(dtype=x.dtype, ndim=x.ndim, layout=x.layout)
            return signature(restype, x)

        return None


@registry.register_global(operator.matmul)
class OperatorMatmulTemplate(AbstractTemplate):
    """Typing template for operator.matmul (@) with arrays"""

    def generic(self, args, kws):
        if len(args) != 2:
            return None

        lhs, rhs = args

        # Only handle array @ array
        if isinstance(lhs, types.Array) and isinstance(rhs, types.Array):
            # Handle 1D @ 1D and 2D @ 2D cases
            if (lhs.ndim == 1 and rhs.ndim == 1) or (lhs.ndim == 2 and rhs.ndim == 2):
                target_dtype = lhs.dtype.unify(self.context, rhs.dtype)
                if target_dtype is None:
                    return None

                restype = types.Array(dtype=target_dtype, ndim=lhs.ndim, layout=lhs.layout)
                return signature(restype, lhs, rhs)

        return None




# Autogenerated by mlir-tblgen; don't manually edit.

from ._ods_common import _cext as _ods_cext
from ._ods_common import (
    equally_sized_accessor as _ods_equally_sized_accessor,
    get_default_loc_context as _ods_get_default_loc_context,
    get_op_results_or_values as _get_op_results_or_values,
    segmented_accessor as _ods_segmented_accessor,
)
_ods_ir = _ods_cext.ir
_ods_cext.globals.register_traceback_file_exclusion(__file__)

import builtins
from typing import Any as _Any, Sequence as _Sequence, Union as _Union, Optional as _Optional
import sys as _sys
if _sys.version_info >= (3, 12):
  from collections.abc import Buffer as _Buffer  # pytype: disable=not-supported-yet
else:
  try:
    from typing_extensions import Buffer as _Buffer
  except ImportError:
    _Buffer = _Any


@_ods_cext.register_dialect
class _Dialect(_ods_ir.Dialect):
  DIALECT_NAMESPACE = "irdl"

@_ods_cext.register_operation(_Dialect)
class AllOfOp(_ods_ir.OpView):
  r"""
  `irdl.all_of` defines a constraint that accepts any type or attribute that
  satisfies all of its provided constraints.
  
  Example:
  
  ```mlir
  irdl.dialect @cmath {
    irdl.type @complex_f32 {
      %0 = irdl.is i32
      %1 = irdl.is f32
      %2 = irdl.any_of(%0, %1) // is 32-bit
  
      %3 = irdl.is f32
      %4 = irdl.is f64
      %5 = irdl.any_of(%3, %4) // is a float
  
      %6 = irdl.all_of(%2, %5) // is a 32-bit float
      irdl.parameters(%6)
    }
  }
  ```
  
  The above program defines a type `complex` inside the dialect `cmath` that
  has one parameter that must be 32-bit long and a float (in other
  words, that must be `f32`).
  """

  OPERATION_NAME = "irdl.all_of"

  _ODS_REGIONS = (0, True)

  def __init__(self, args: _Sequence[_ods_ir.Value], *, results: _Optional[_Sequence[_ods_ir.Type]] = None, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None):
    operands = []
    attributes = {}
    regions = None
    operands.extend(_get_op_results_or_values(args))
    _ods_context = _ods_get_default_loc_context(loc)
    if results is None: results = [operands[0].type] * 1
    _ods_successors = None
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

  @builtins.property
  def args(self) -> _ods_ir.OpOperandList:
    _ods_variadic_group_length = len(self.operation.operands) - 1 + 1
    return self.operation.operands[0:0 + _ods_variadic_group_length]

  @builtins.property
  def output(self) -> _ods_ir.OpResult:
    return self.operation.results[0]

@_ods_cext.register_op_adaptor(AllOfOp)
class AllOfOpAdaptor(_ods_ir.OpAdaptor):
  OPERATION_NAME = "irdl.all_of"

  @builtins.property
  def args(self) -> _ods_ir.OpOperandList:
    _ods_variadic_group_length = len(self.operands) - 1 + 1
    return self.operands[0:0 + _ods_variadic_group_length]

def all_of(args: _Sequence[_ods_ir.Value], *, results: _Optional[_Sequence[_ods_ir.Type]] = None, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None) -> _ods_ir.OpResult:
  return AllOfOp(args=args, results=results, loc=loc, ip=ip).result

@_ods_cext.register_operation(_Dialect)
class AnyOfOp(_ods_ir.OpView):
  r"""
  `irdl.any_of` defines a constraint that accepts any type or attribute that
  satisfies at least one of its provided type constraints.
  
  Example:
  
  ```mlir
  irdl.dialect @cmath {
    irdl.type @complex {
      %0 = irdl.is i32
      %1 = irdl.is i64
      %2 = irdl.is f32
      %3 = irdl.is f64
      %4 = irdl.any_of(%0, %1, %2, %3)
      irdl.parameters(%4)
    }
  }
  ```
  
  The above program defines a type `complex` inside the dialect `cmath` that
  has a single type parameter that can be either `i32`, `i64`, `f32` or
  `f64`.
  """

  OPERATION_NAME = "irdl.any_of"

  _ODS_REGIONS = (0, True)

  def __init__(self, args: _Sequence[_ods_ir.Value], *, results: _Optional[_Sequence[_ods_ir.Type]] = None, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None):
    operands = []
    attributes = {}
    regions = None
    operands.extend(_get_op_results_or_values(args))
    _ods_context = _ods_get_default_loc_context(loc)
    if results is None: results = [operands[0].type] * 1
    _ods_successors = None
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

  @builtins.property
  def args(self) -> _ods_ir.OpOperandList:
    _ods_variadic_group_length = len(self.operation.operands) - 1 + 1
    return self.operation.operands[0:0 + _ods_variadic_group_length]

  @builtins.property
  def output(self) -> _ods_ir.OpResult:
    return self.operation.results[0]

@_ods_cext.register_op_adaptor(AnyOfOp)
class AnyOfOpAdaptor(_ods_ir.OpAdaptor):
  OPERATION_NAME = "irdl.any_of"

  @builtins.property
  def args(self) -> _ods_ir.OpOperandList:
    _ods_variadic_group_length = len(self.operands) - 1 + 1
    return self.operands[0:0 + _ods_variadic_group_length]

def any_of(args: _Sequence[_ods_ir.Value], *, results: _Optional[_Sequence[_ods_ir.Type]] = None, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None) -> _ods_ir.OpResult:
  return AnyOfOp(args=args, results=results, loc=loc, ip=ip).result

@_ods_cext.register_operation(_Dialect)
class AnyOp(_ods_ir.OpView):
  r"""
  `irdl.any` defines a constraint that accepts any type or attribute.
  
  Example:
  
  ```mlir
  irdl.dialect @cmath {
    irdl.type @complex_flexible {
      %0 = irdl.any
      irdl.parameters(%0)
    }
  }
  ```
  
  The above program defines a type `complex_flexible` inside the dialect
  `cmath` that has a single parameter that can be any attribute.
  """

  OPERATION_NAME = "irdl.any"

  _ODS_REGIONS = (0, True)

  def __init__(self, *, results: _Optional[_Sequence[_ods_ir.Type]] = None, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None):
    operands = []
    attributes = {}
    regions = None
    _ods_context = _ods_get_default_loc_context(loc)
    _ods_successors = None
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

  @builtins.property
  def output(self) -> _ods_ir.OpResult:
    return self.operation.results[0]

@_ods_cext.register_op_adaptor(AnyOp)
class AnyOpAdaptor(_ods_ir.OpAdaptor):
  OPERATION_NAME = "irdl.any"

def any(*, results: _Optional[_Sequence[_ods_ir.Type]] = None, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None) -> _ods_ir.OpResult:
  return AnyOp(results=results, loc=loc, ip=ip).result

@_ods_cext.register_operation(_Dialect)
class AttributeOp(_ods_ir.OpView):
  r"""
  `irdl.attribute` defines a new attribute belonging to the `irdl.dialect`
  parent.
  
  The attribute parameters can be defined with an `irdl.parameters` operation
  in the optional region.
  
  Example:
  
  ```mlir
  irdl.dialect @testd {
    irdl.attribute @enum_attr {
      %0 = irdl.is "foo"
      %1 = irdl.is "bar"
      %2 = irdl.any_of(%0, %1)
      irdl.parameters(%2)
    }
  }
  ```
  
  The above program defines an `enum_attr` attribute inside the `testd`
  dialect. The attribute has one `StringAttr` parameter that should be
  either a `"foo"` or a `"bar"`.
  """

  OPERATION_NAME = "irdl.attribute"

  _ODS_REGIONS = (1, True)

  def __init__(self, sym_name: _Union[str, _ods_ir.StringAttr], *, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None):
    operands = []
    attributes = {}
    regions = None
    _ods_context = _ods_get_default_loc_context(loc)
    attributes["sym_name"] = (sym_name if (
    isinstance(sym_name, _ods_ir.Attribute) or
    not _ods_ir.AttrBuilder.contains('SymbolNameAttr')) else
      _ods_ir.AttrBuilder.get('SymbolNameAttr')(sym_name, context=_ods_context))
    results = []
    _ods_successors = None
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

  @builtins.property
  def sym_name(self) -> _ods_ir.StringAttr:
    return self.operation.attributes["sym_name"]

  @sym_name.setter
  def sym_name(self, value: _ods_ir.StringAttr):
    if value is None:
      raise ValueError("'None' not allowed as value for mandatory attributes")
    self.operation.attributes["sym_name"] = value

  @builtins.property
  def body(self) -> _ods_ir.Region:
    return self.regions[0]

@_ods_cext.register_op_adaptor(AttributeOp)
class AttributeOpAdaptor(_ods_ir.OpAdaptor):
  OPERATION_NAME = "irdl.attribute"

  @builtins.property
  def sym_name(self) -> _ods_ir.StringAttr:
    return self.attributes["sym_name"]

def attribute(sym_name: _Union[str, _ods_ir.StringAttr], *, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None) -> AttributeOp:
  return AttributeOp(sym_name=sym_name, loc=loc, ip=ip)

@_ods_cext.register_operation(_Dialect)
class AttributesOp(_ods_ir.OpView):
  r"""
  `irdl.attributes` defines the attributes of the `irdl.operation` parent
  operation definition.
  
  In the following example, `irdl.attributes` defines the attributes of the
  `attr_op` operation:
  
  ```mlir
  irdl.dialect @example {
  
    irdl.operation @attr_op {
      %0 = irdl.any
      %1 = irdl.is i64
      irdl.attibutes {
        "attr1" = %0,
        "attr2" = %1
      }
    }
  }
  ```
  
  The operation will expect an arbitrary attribute "attr1" and an
  attribute "attr2" with value `i64`.
  """

  OPERATION_NAME = "irdl.attributes"

  _ODS_REGIONS = (0, True)

  def __init__(self, attributeValues: _Sequence[_ods_ir.Value], attributeValueNames: _Union[_Sequence[str], _ods_ir.ArrayAttr], *, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None):
    operands = []
    attributes = {}
    regions = None
    operands.extend(_get_op_results_or_values(attributeValues))
    _ods_context = _ods_get_default_loc_context(loc)
    attributes["attributeValueNames"] = (attributeValueNames if (
    isinstance(attributeValueNames, _ods_ir.Attribute) or
    not _ods_ir.AttrBuilder.contains('StrArrayAttr')) else
      _ods_ir.AttrBuilder.get('StrArrayAttr')(attributeValueNames, context=_ods_context))
    results = []
    _ods_successors = None
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

  @builtins.property
  def attributeValues(self) -> _ods_ir.OpOperandList:
    _ods_variadic_group_length = len(self.operation.operands) - 1 + 1
    return self.operation.operands[0:0 + _ods_variadic_group_length]

  @builtins.property
  def attributeValueNames(self) -> _ods_ir.ArrayAttr:
    return self.operation.attributes["attributeValueNames"]

  @attributeValueNames.setter
  def attributeValueNames(self, value: _ods_ir.ArrayAttr):
    if value is None:
      raise ValueError("'None' not allowed as value for mandatory attributes")
    self.operation.attributes["attributeValueNames"] = value

@_ods_cext.register_op_adaptor(AttributesOp)
class AttributesOpAdaptor(_ods_ir.OpAdaptor):
  OPERATION_NAME = "irdl.attributes"

  @builtins.property
  def attributeValues(self) -> _ods_ir.OpOperandList:
    _ods_variadic_group_length = len(self.operands) - 1 + 1
    return self.operands[0:0 + _ods_variadic_group_length]

  @builtins.property
  def attributeValueNames(self) -> _ods_ir.ArrayAttr:
    return self.attributes["attributeValueNames"]

def attributes_(attribute_values: _Sequence[_ods_ir.Value], attribute_value_names: _Union[_Sequence[str], _ods_ir.ArrayAttr], *, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None) -> AttributesOp:
  return AttributesOp(attributeValues=attribute_values, attributeValueNames=attribute_value_names, loc=loc, ip=ip)

@_ods_cext.register_operation(_Dialect)
class BaseOp(_ods_ir.OpView):
  r"""
  `irdl.base` defines a constraint that only accepts a single type
  or attribute base, e.g. an `IntegerType`. The attribute base is defined
  either by a symbolic reference to the corresponding IRDL definition,
  or by the name of the base. Named bases are prefixed with `!` or `#`
  respectively for types and attributes.
  
  Example:
  
  ```mlir
  irdl.dialect @cmath {
    irdl.type @complex {
      %0 = irdl.base "!builtin.integer"
      irdl.parameters(%0)
    }
  
    irdl.type @complex_wrapper {
      %0 = irdl.base @cmath::@complex
      irdl.parameters(%0)
    }
  }
  ```
  
  The above program defines a `cmath.complex` type that expects a single
  parameter, which is a type with base name `builtin.integer`, which is the
  name of an `IntegerType` type.
  It also defines a `cmath.complex_wrapper` type that expects a single
  parameter, which is a type of base type `cmath.complex`.
  """

  OPERATION_NAME = "irdl.base"

  _ODS_REGIONS = (0, True)

  def __init__(self, *, base_ref: _Optional[_Union[str, _ods_ir.SymbolRefAttr]] = None, base_name: _Optional[_Union[str, _ods_ir.StringAttr]] = None, results: _Optional[_Sequence[_ods_ir.Type]] = None, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None):
    operands = []
    attributes = {}
    regions = None
    _ods_context = _ods_get_default_loc_context(loc)
    if base_ref is not None: attributes["base_ref"] = (base_ref if (
        isinstance(base_ref, _ods_ir.Attribute) or
        not _ods_ir.AttrBuilder.contains('SymbolRefAttr')) else
          _ods_ir.AttrBuilder.get('SymbolRefAttr')(base_ref, context=_ods_context))
    if base_name is not None: attributes["base_name"] = (base_name if (
        isinstance(base_name, _ods_ir.Attribute) or
        not _ods_ir.AttrBuilder.contains('StrAttr')) else
          _ods_ir.AttrBuilder.get('StrAttr')(base_name, context=_ods_context))
    _ods_successors = None
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

  @builtins.property
  def base_ref(self) -> _Optional[_ods_ir.SymbolRefAttr]:
    if "base_ref" not in self.operation.attributes:
      return None
    return self.operation.attributes["base_ref"]

  @base_ref.setter
  def base_ref(self, value: _Optional[_ods_ir.SymbolRefAttr]):
    if value is not None:
      self.operation.attributes["base_ref"] = value
    elif "base_ref" in self.operation.attributes:
      del self.operation.attributes["base_ref"]

  @base_ref.deleter
  def base_ref(self):
    del self.operation.attributes["base_ref"]

  @builtins.property
  def base_name(self) -> _Optional[_ods_ir.StringAttr]:
    if "base_name" not in self.operation.attributes:
      return None
    return self.operation.attributes["base_name"]

  @base_name.setter
  def base_name(self, value: _Optional[_ods_ir.StringAttr]):
    if value is not None:
      self.operation.attributes["base_name"] = value
    elif "base_name" in self.operation.attributes:
      del self.operation.attributes["base_name"]

  @base_name.deleter
  def base_name(self):
    del self.operation.attributes["base_name"]

  @builtins.property
  def output(self) -> _ods_ir.OpResult:
    return self.operation.results[0]

@_ods_cext.register_op_adaptor(BaseOp)
class BaseOpAdaptor(_ods_ir.OpAdaptor):
  OPERATION_NAME = "irdl.base"

  @builtins.property
  def base_ref(self) -> _Optional[_ods_ir.SymbolRefAttr]:
    if "base_ref" not in self.attributes:
      return None
    return self.attributes["base_ref"]

  @builtins.property
  def base_name(self) -> _Optional[_ods_ir.StringAttr]:
    if "base_name" not in self.attributes:
      return None
    return self.attributes["base_name"]

def base(*, base_ref: _Optional[_Union[str, _ods_ir.SymbolRefAttr]] = None, base_name: _Optional[_Union[str, _ods_ir.StringAttr]] = None, results: _Optional[_Sequence[_ods_ir.Type]] = None, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None) -> _ods_ir.OpResult:
  return BaseOp(base_ref=base_ref, base_name=base_name, results=results, loc=loc, ip=ip).result

@_ods_cext.register_operation(_Dialect)
class CPredOp(_ods_ir.OpView):
  r"""
  `irdl.c_pred` defines a constraint that is written in C++.
  
  Dialects using this operation cannot be registered at runtime, as it relies
  on C++ code.
  
  Special placeholders can be used to refer to entities in the context where
  this predicate is used. They serve as "hooks" to the enclosing environment.
  The following special placeholders are supported in constraints for an op:
  
  * `$_builder` will be replaced by a mlir::Builder instance.
  * `$_op` will be replaced by the current operation.
  * `$_self` will be replaced with the entity this predicate is attached to.
     Compared to ODS, `$_self` is always of type `mlir::Attribute`, and types
     are manipulated as `TypeAttr` attributes.
  
  Example:
  ```mlir
  irdl.type @op_with_attr {
    %0 = irdl.c_pred "::llvm::isa<::mlir::IntegerAttr>($_self)"
    irdl.parameters(%0)
  }
  ```
  
  In this example, @op_with_attr is defined as a type with a single
  parameter, which is an `IntegerAttr`, as constrained by the C++ predicate.
  """

  OPERATION_NAME = "irdl.c_pred"

  _ODS_REGIONS = (0, True)

  def __init__(self, pred: _Union[str, _ods_ir.StringAttr], *, results: _Optional[_Sequence[_ods_ir.Type]] = None, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None):
    operands = []
    attributes = {}
    regions = None
    _ods_context = _ods_get_default_loc_context(loc)
    attributes["pred"] = (pred if (
    isinstance(pred, _ods_ir.Attribute) or
    not _ods_ir.AttrBuilder.contains('StrAttr')) else
      _ods_ir.AttrBuilder.get('StrAttr')(pred, context=_ods_context))
    _ods_successors = None
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

  @builtins.property
  def pred(self) -> _ods_ir.StringAttr:
    return self.operation.attributes["pred"]

  @pred.setter
  def pred(self, value: _ods_ir.StringAttr):
    if value is None:
      raise ValueError("'None' not allowed as value for mandatory attributes")
    self.operation.attributes["pred"] = value

  @builtins.property
  def output(self) -> _ods_ir.OpResult:
    return self.operation.results[0]

@_ods_cext.register_op_adaptor(CPredOp)
class CPredOpAdaptor(_ods_ir.OpAdaptor):
  OPERATION_NAME = "irdl.c_pred"

  @builtins.property
  def pred(self) -> _ods_ir.StringAttr:
    return self.attributes["pred"]

def c_pred(pred: _Union[str, _ods_ir.StringAttr], *, results: _Optional[_Sequence[_ods_ir.Type]] = None, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None) -> _ods_ir.OpResult:
  return CPredOp(pred=pred, results=results, loc=loc, ip=ip).result

@_ods_cext.register_operation(_Dialect)
class DialectOp(_ods_ir.OpView):
  r"""
  The `irdl.dialect` operation defines a dialect. All operations, attributes,
  and types defined inside its region will be part of the dialect.
  
  Example:
  
  ```mlir
  irdl.dialect @cmath {
    ...
  }
  ```
  
  The above program defines a `cmath` dialect.
  """

  OPERATION_NAME = "irdl.dialect"

  _ODS_REGIONS = (1, True)

  def __init__(self, sym_name: _Union[str, _ods_ir.StringAttr], *, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None):
    operands = []
    attributes = {}
    regions = None
    _ods_context = _ods_get_default_loc_context(loc)
    attributes["sym_name"] = (sym_name if (
    isinstance(sym_name, _ods_ir.Attribute) or
    not _ods_ir.AttrBuilder.contains('SymbolNameAttr')) else
      _ods_ir.AttrBuilder.get('SymbolNameAttr')(sym_name, context=_ods_context))
    results = []
    _ods_successors = None
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

  @builtins.property
  def sym_name(self) -> _ods_ir.StringAttr:
    return self.operation.attributes["sym_name"]

  @sym_name.setter
  def sym_name(self, value: _ods_ir.StringAttr):
    if value is None:
      raise ValueError("'None' not allowed as value for mandatory attributes")
    self.operation.attributes["sym_name"] = value

  @builtins.property
  def body(self) -> _ods_ir.Region:
    return self.regions[0]

@_ods_cext.register_op_adaptor(DialectOp)
class DialectOpAdaptor(_ods_ir.OpAdaptor):
  OPERATION_NAME = "irdl.dialect"

  @builtins.property
  def sym_name(self) -> _ods_ir.StringAttr:
    return self.attributes["sym_name"]

def dialect(sym_name: _Union[str, _ods_ir.StringAttr], *, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None) -> DialectOp:
  return DialectOp(sym_name=sym_name, loc=loc, ip=ip)

@_ods_cext.register_operation(_Dialect)
class IsOp(_ods_ir.OpView):
  r"""
  `irdl.is` defines a constraint that only accepts a specific instance of a
  type or attribute.
  
  Example:
  
  ```mlir
  irdl.dialect @cmath {
    irdl.type @complex_i32 {
      %0 = irdl.is i32
      irdl.parameters(%0)
    }
  }
  ```
  
  The above program defines a `complex_i32` type inside the dialect `cmath`
  that can only have a `i32` as its parameter.
  """

  OPERATION_NAME = "irdl.is"

  _ODS_REGIONS = (0, True)

  def __init__(self, expected: _Union[_Any, _ods_ir.Attribute], *, results: _Optional[_Sequence[_ods_ir.Type]] = None, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None):
    operands = []
    attributes = {}
    regions = None
    _ods_context = _ods_get_default_loc_context(loc)
    attributes["expected"] = (expected if (
    isinstance(expected, _ods_ir.Attribute) or
    not _ods_ir.AttrBuilder.contains('AnyAttr')) else
      _ods_ir.AttrBuilder.get('AnyAttr')(expected, context=_ods_context))
    _ods_successors = None
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

  @builtins.property
  def expected(self) -> _ods_ir.Attribute:
    return self.operation.attributes["expected"]

  @expected.setter
  def expected(self, value: _ods_ir.Attribute):
    if value is None:
      raise ValueError("'None' not allowed as value for mandatory attributes")
    self.operation.attributes["expected"] = value

  @builtins.property
  def output(self) -> _ods_ir.OpResult:
    return self.operation.results[0]

@_ods_cext.register_op_adaptor(IsOp)
class IsOpAdaptor(_ods_ir.OpAdaptor):
  OPERATION_NAME = "irdl.is"

  @builtins.property
  def expected(self) -> _ods_ir.Attribute:
    return self.attributes["expected"]

def is_(expected: _Union[_Any, _ods_ir.Attribute], *, results: _Optional[_Sequence[_ods_ir.Type]] = None, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None) -> _ods_ir.OpResult:
  return IsOp(expected=expected, results=results, loc=loc, ip=ip).result

@_ods_cext.register_operation(_Dialect)
class OperandsOp(_ods_ir.OpView):
  r"""
  `irdl.operands` define the operands of the `irdl.operation` parent operation
  definition. Each operand is named after an identifier.
  
  In the following example, `irdl.operands` defines the operands of the
  `mul` operation:
  
  ```mlir
  irdl.dialect @cmath {
  
    irdl.type @complex { /* ... */ }
  
    irdl.operation @mul {
      %0 = irdl.any
      %1 = irdl.parametric @cmath::@complex<%0>
      irdl.results(res: %1)
      irdl.operands(lhs: %1, rhs: %1)
    }
  }
  ```
  
  The `mul` operation will expect two operands of type `cmath.complex`, that
  have the same type, and return a result of the same type.
  
  The operands can also be marked as variadic or optional:
  ```mlir
  irdl.operands(foo: %0, bar: single %1, baz: optional %2, qux: variadic %3)
  ```
  
  Here, foo and bar are required single operands, baz is an optional operand,
  and qux is a variadic operand.
  
  When more than one operand is marked as optional or variadic, the operation
  will expect a 'operandSegmentSizes' attribute that defines the number of
  operands in each segment.
  """

  OPERATION_NAME = "irdl.operands"

  _ODS_REGIONS = (0, True)

  def __init__(self, args: _Sequence[_ods_ir.Value], names: _Union[_Sequence[str], _ods_ir.ArrayAttr], variadicity: _Union[_Any, _ods_ir.Attribute], *, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None):
    operands = []
    attributes = {}
    regions = None
    operands.extend(_get_op_results_or_values(args))
    _ods_context = _ods_get_default_loc_context(loc)
    attributes["names"] = (names if (
    isinstance(names, _ods_ir.Attribute) or
    not _ods_ir.AttrBuilder.contains('StrArrayAttr')) else
      _ods_ir.AttrBuilder.get('StrArrayAttr')(names, context=_ods_context))
    attributes["variadicity"] = (variadicity if (
    isinstance(variadicity, _ods_ir.Attribute) or
    not _ods_ir.AttrBuilder.contains('VariadicityArrayAttr')) else
      _ods_ir.AttrBuilder.get('VariadicityArrayAttr')(variadicity, context=_ods_context))
    results = []
    _ods_successors = None
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

  @builtins.property
  def args(self) -> _ods_ir.OpOperandList:
    _ods_variadic_group_length = len(self.operation.operands) - 1 + 1
    return self.operation.operands[0:0 + _ods_variadic_group_length]

  @builtins.property
  def names(self) -> _ods_ir.ArrayAttr:
    return self.operation.attributes["names"]

  @names.setter
  def names(self, value: _ods_ir.ArrayAttr):
    if value is None:
      raise ValueError("'None' not allowed as value for mandatory attributes")
    self.operation.attributes["names"] = value

  @builtins.property
  def variadicity(self) -> _ods_ir.Attribute:
    return self.operation.attributes["variadicity"]

  @variadicity.setter
  def variadicity(self, value: _ods_ir.Attribute):
    if value is None:
      raise ValueError("'None' not allowed as value for mandatory attributes")
    self.operation.attributes["variadicity"] = value

@_ods_cext.register_op_adaptor(OperandsOp)
class OperandsOpAdaptor(_ods_ir.OpAdaptor):
  OPERATION_NAME = "irdl.operands"

  @builtins.property
  def args(self) -> _ods_ir.OpOperandList:
    _ods_variadic_group_length = len(self.operands) - 1 + 1
    return self.operands[0:0 + _ods_variadic_group_length]

  @builtins.property
  def names(self) -> _ods_ir.ArrayAttr:
    return self.attributes["names"]

  @builtins.property
  def variadicity(self) -> _ods_ir.Attribute:
    return self.attributes["variadicity"]

def operands_(args: _Sequence[_ods_ir.Value], names: _Union[_Sequence[str], _ods_ir.ArrayAttr], variadicity: _Union[_Any, _ods_ir.Attribute], *, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None) -> OperandsOp:
  return OperandsOp(args=args, names=names, variadicity=variadicity, loc=loc, ip=ip)

@_ods_cext.register_operation(_Dialect)
class OperationOp(_ods_ir.OpView):
  r"""
  `irdl.operation` defines a new operation belonging to the `irdl.dialect`
  parent.
  
  Operations can define constraints on their operands and results with the
  `irdl.results` and `irdl.operands` operations. If these operations are not
  present in the region, the results or operands are expected to be empty.
  
  Example:
  
  ```mlir
  irdl.dialect @cmath {
  
    irdl.type @complex { /* ... */ }
  
    irdl.operation @norm {
      %0 = irdl.any
      %1 = irdl.parametric @cmath::@complex<%0>
      irdl.results(%0)
      irdl.operands(%1)
    }
  }
  ```
  
  The above program defines an operation `norm` inside the dialect `cmath`.
  The operation expects a single operand of base type `cmath.complex`, and
  returns a single result of the element type of the operand.
  """

  OPERATION_NAME = "irdl.operation"

  _ODS_REGIONS = (1, True)

  def __init__(self, sym_name: _Union[str, _ods_ir.StringAttr], *, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None):
    operands = []
    attributes = {}
    regions = None
    _ods_context = _ods_get_default_loc_context(loc)
    attributes["sym_name"] = (sym_name if (
    isinstance(sym_name, _ods_ir.Attribute) or
    not _ods_ir.AttrBuilder.contains('SymbolNameAttr')) else
      _ods_ir.AttrBuilder.get('SymbolNameAttr')(sym_name, context=_ods_context))
    results = []
    _ods_successors = None
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

  @builtins.property
  def sym_name(self) -> _ods_ir.StringAttr:
    return self.operation.attributes["sym_name"]

  @sym_name.setter
  def sym_name(self, value: _ods_ir.StringAttr):
    if value is None:
      raise ValueError("'None' not allowed as value for mandatory attributes")
    self.operation.attributes["sym_name"] = value

  @builtins.property
  def body(self) -> _ods_ir.Region:
    return self.regions[0]

@_ods_cext.register_op_adaptor(OperationOp)
class OperationOpAdaptor(_ods_ir.OpAdaptor):
  OPERATION_NAME = "irdl.operation"

  @builtins.property
  def sym_name(self) -> _ods_ir.StringAttr:
    return self.attributes["sym_name"]

def operation_(sym_name: _Union[str, _ods_ir.StringAttr], *, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None) -> OperationOp:
  return OperationOp(sym_name=sym_name, loc=loc, ip=ip)

@_ods_cext.register_operation(_Dialect)
class ParametersOp(_ods_ir.OpView):
  r"""
  `irdl.parameters` defines the constraints on parameters of a type or
  attribute definition. Each parameter is named after an identifier.
  
  Example:
  
  ```mlir
  irdl.dialect @cmath {
    irdl.type @complex {
      %0 = irdl.is i32
      %1 = irdl.is i64
      %2 = irdl.any_of(%0, %1)
      irdl.parameters(elem: %2)
    }
  }
  ```
  
  The above program defines a type `complex` inside the dialect `cmath`. The
  type has a single parameter `elem` that should be either `i32` or `i64`.
  """

  OPERATION_NAME = "irdl.parameters"

  _ODS_REGIONS = (0, True)

  def __init__(self, args: _Sequence[_ods_ir.Value], names: _Union[_Sequence[str], _ods_ir.ArrayAttr], *, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None):
    operands = []
    attributes = {}
    regions = None
    operands.extend(_get_op_results_or_values(args))
    _ods_context = _ods_get_default_loc_context(loc)
    attributes["names"] = (names if (
    isinstance(names, _ods_ir.Attribute) or
    not _ods_ir.AttrBuilder.contains('StrArrayAttr')) else
      _ods_ir.AttrBuilder.get('StrArrayAttr')(names, context=_ods_context))
    results = []
    _ods_successors = None
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

  @builtins.property
  def args(self) -> _ods_ir.OpOperandList:
    _ods_variadic_group_length = len(self.operation.operands) - 1 + 1
    return self.operation.operands[0:0 + _ods_variadic_group_length]

  @builtins.property
  def names(self) -> _ods_ir.ArrayAttr:
    return self.operation.attributes["names"]

  @names.setter
  def names(self, value: _ods_ir.ArrayAttr):
    if value is None:
      raise ValueError("'None' not allowed as value for mandatory attributes")
    self.operation.attributes["names"] = value

@_ods_cext.register_op_adaptor(ParametersOp)
class ParametersOpAdaptor(_ods_ir.OpAdaptor):
  OPERATION_NAME = "irdl.parameters"

  @builtins.property
  def args(self) -> _ods_ir.OpOperandList:
    _ods_variadic_group_length = len(self.operands) - 1 + 1
    return self.operands[0:0 + _ods_variadic_group_length]

  @builtins.property
  def names(self) -> _ods_ir.ArrayAttr:
    return self.attributes["names"]

def parameters(args: _Sequence[_ods_ir.Value], names: _Union[_Sequence[str], _ods_ir.ArrayAttr], *, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None) -> ParametersOp:
  return ParametersOp(args=args, names=names, loc=loc, ip=ip)

@_ods_cext.register_operation(_Dialect)
class ParametricOp(_ods_ir.OpView):
  r"""
  `irdl.parametric` defines a constraint that accepts only a single type
  or attribute base. The attribute base is defined by a symbolic reference
  to the corresponding definition. It will additionally constraint the
  parameters of the type/attribute.
  
  Example:
  
  ```mlir
  irdl.dialect @cmath {
  
    irdl.type @complex { /* ... */ }
  
    irdl.operation @norm {
      %0 = irdl.any
      %1 = irdl.parametric @cmath::@complex<%0>
      irdl.operands(%1)
      irdl.results(%0)
    }
  }
  ```
  
  The above program defines an operation `norm` inside the dialect `cmath` that
  for any `T` takes a `cmath.complex` with parameter `T` and returns a `T`.
  """

  OPERATION_NAME = "irdl.parametric"

  _ODS_REGIONS = (0, True)

  def __init__(self, base_type: _Union[str, _ods_ir.SymbolRefAttr], args: _Sequence[_ods_ir.Value], *, results: _Optional[_Sequence[_ods_ir.Type]] = None, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None):
    operands = []
    attributes = {}
    regions = None
    operands.extend(_get_op_results_or_values(args))
    _ods_context = _ods_get_default_loc_context(loc)
    attributes["base_type"] = (base_type if (
    isinstance(base_type, _ods_ir.Attribute) or
    not _ods_ir.AttrBuilder.contains('SymbolRefAttr')) else
      _ods_ir.AttrBuilder.get('SymbolRefAttr')(base_type, context=_ods_context))
    _ods_successors = None
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

  @builtins.property
  def args(self) -> _ods_ir.OpOperandList:
    _ods_variadic_group_length = len(self.operation.operands) - 1 + 1
    return self.operation.operands[0:0 + _ods_variadic_group_length]

  @builtins.property
  def base_type(self) -> _ods_ir.SymbolRefAttr:
    return self.operation.attributes["base_type"]

  @base_type.setter
  def base_type(self, value: _ods_ir.SymbolRefAttr):
    if value is None:
      raise ValueError("'None' not allowed as value for mandatory attributes")
    self.operation.attributes["base_type"] = value

  @builtins.property
  def output(self) -> _ods_ir.OpResult:
    return self.operation.results[0]

@_ods_cext.register_op_adaptor(ParametricOp)
class ParametricOpAdaptor(_ods_ir.OpAdaptor):
  OPERATION_NAME = "irdl.parametric"

  @builtins.property
  def args(self) -> _ods_ir.OpOperandList:
    _ods_variadic_group_length = len(self.operands) - 1 + 1
    return self.operands[0:0 + _ods_variadic_group_length]

  @builtins.property
  def base_type(self) -> _ods_ir.SymbolRefAttr:
    return self.attributes["base_type"]

def parametric(base_type: _Union[str, _ods_ir.SymbolRefAttr], args: _Sequence[_ods_ir.Value], *, results: _Optional[_Sequence[_ods_ir.Type]] = None, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None) -> _ods_ir.OpResult:
  return ParametricOp(base_type=base_type, args=args, results=results, loc=loc, ip=ip).result

@_ods_cext.register_operation(_Dialect)
class RegionOp(_ods_ir.OpView):
  r"""
  The irdl.region construct defines a set of characteristics
  that a region of an operation should satify. Each region is named after
  an identifier.
  
  These characteristics include constraints for the entry block arguments
  of the region and the total number of blocks it contains.
  The number of blocks must be a non-zero and non-negative integer,
  and it is optional by default.
  The set of constraints for the entry block arguments may be optional or
  empty. If no parentheses are provided, the set is assumed to be optional,
  and the arguments are not constrained in any way. If parentheses are
  provided with no arguments, it means that the region must have
  no entry block arguments
  
  
  Example:
  
  ```mlir
  irdl.dialect @example {
    irdl.operation @op_with_regions {
        %r0 = irdl.region
        %r1 = irdl.region()
        %v0 = irdl.is i32
        %v1 = irdl.is i64
        %r2 = irdl.region(%v0, %v1)
        %r3 = irdl.region with size 3
  
        irdl.regions(foo: %r0, bar: %r1, baz: %r2, qux: %r3)
    }
  }
  ```
  
  The above snippet demonstrates an operation named `@op_with_regions`,
  which is constrained to have four regions.
  
  * Region `foo` doesn't have any constraints on the arguments
    or the number of blocks.
  * Region `bar` should have an empty set of arguments.
  * Region `baz` should have two arguments of types `i32` and `i64`.
  * Region `qux` should contain exactly three blocks.
  """

  OPERATION_NAME = "irdl.region"

  _ODS_REGIONS = (0, True)

  def __init__(self, entryBlockArgs: _Sequence[_ods_ir.Value], *, numberOfBlocks: _Optional[_Union[int, _ods_ir.IntegerAttr]] = None, constrainedArguments: _Optional[bool] = None, results: _Optional[_Sequence[_ods_ir.Type]] = None, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None):
    operands = []
    attributes = {}
    regions = None
    operands.extend(_get_op_results_or_values(entryBlockArgs))
    _ods_context = _ods_get_default_loc_context(loc)
    if numberOfBlocks is not None: attributes["numberOfBlocks"] = (numberOfBlocks if (
        isinstance(numberOfBlocks, _ods_ir.Attribute) or
        not _ods_ir.AttrBuilder.contains('I32Attr')) else
          _ods_ir.AttrBuilder.get('I32Attr')(numberOfBlocks, context=_ods_context))
    if bool(constrainedArguments): attributes["constrainedArguments"] = _ods_ir.UnitAttr.get(
      _ods_get_default_loc_context(loc))
    _ods_successors = None
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

  @builtins.property
  def entryBlockArgs(self) -> _ods_ir.OpOperandList:
    _ods_variadic_group_length = len(self.operation.operands) - 1 + 1
    return self.operation.operands[0:0 + _ods_variadic_group_length]

  @builtins.property
  def numberOfBlocks(self) -> _Optional[_ods_ir.IntegerAttr]:
    if "numberOfBlocks" not in self.operation.attributes:
      return None
    return self.operation.attributes["numberOfBlocks"]

  @numberOfBlocks.setter
  def numberOfBlocks(self, value: _Optional[_ods_ir.IntegerAttr]):
    if value is not None:
      self.operation.attributes["numberOfBlocks"] = value
    elif "numberOfBlocks" in self.operation.attributes:
      del self.operation.attributes["numberOfBlocks"]

  @numberOfBlocks.deleter
  def numberOfBlocks(self):
    del self.operation.attributes["numberOfBlocks"]

  @builtins.property
  def constrainedArguments(self) -> bool:
    return "constrainedArguments" in self.operation.attributes

  @constrainedArguments.setter
  def constrainedArguments(self, value):
    if bool(value):
      self.operation.attributes["constrainedArguments"] = _ods_ir.UnitAttr.get()
    elif "constrainedArguments" in self.operation.attributes:
      del self.operation.attributes["constrainedArguments"]

  @constrainedArguments.deleter
  def constrainedArguments(self):
    del self.operation.attributes["constrainedArguments"]

  @builtins.property
  def output(self) -> _ods_ir.OpResult:
    return self.operation.results[0]

@_ods_cext.register_op_adaptor(RegionOp)
class RegionOpAdaptor(_ods_ir.OpAdaptor):
  OPERATION_NAME = "irdl.region"

  @builtins.property
  def entryBlockArgs(self) -> _ods_ir.OpOperandList:
    _ods_variadic_group_length = len(self.operands) - 1 + 1
    return self.operands[0:0 + _ods_variadic_group_length]

  @builtins.property
  def numberOfBlocks(self) -> _Optional[_ods_ir.IntegerAttr]:
    if "numberOfBlocks" not in self.attributes:
      return None
    return self.attributes["numberOfBlocks"]

  @builtins.property
  def constrainedArguments(self) -> bool:
    return "constrainedArguments" in self.attributes

def region(entry_block_args: _Sequence[_ods_ir.Value], *, number_of_blocks: _Optional[_Union[int, _ods_ir.IntegerAttr]] = None, constrained_arguments: _Optional[bool] = None, results: _Optional[_Sequence[_ods_ir.Type]] = None, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None) -> _ods_ir.OpResult:
  return RegionOp(entryBlockArgs=entry_block_args, numberOfBlocks=number_of_blocks, constrainedArguments=constrained_arguments, results=results, loc=loc, ip=ip).result

@_ods_cext.register_operation(_Dialect)
class RegionsOp(_ods_ir.OpView):
  r"""
  `irdl.regions` defines the regions of an operation by accepting
  values produced by `irdl.region` operation as arguments. Each
  region has an identifier as name.
  
  Example:
  
  ```mlir
  irdl.dialect @example {
    irdl.operation @op_with_regions {
      %r1 = irdl.region with size 3
      %0 = irdl.any
      %r2 = irdl.region(%0)
      irdl.regions(foo: %r1, bar: %r2)
    }
  }
  ```
  
  In the snippet above the operation is constrained to have two regions.
  The first region (`foo`) should contain three blocks.
  The second region (`bar`) should have one region with one argument.
  """

  OPERATION_NAME = "irdl.regions"

  _ODS_REGIONS = (0, True)

  def __init__(self, args: _Sequence[_ods_ir.Value], names: _Union[_Sequence[str], _ods_ir.ArrayAttr], *, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None):
    operands = []
    attributes = {}
    regions = None
    operands.extend(_get_op_results_or_values(args))
    _ods_context = _ods_get_default_loc_context(loc)
    attributes["names"] = (names if (
    isinstance(names, _ods_ir.Attribute) or
    not _ods_ir.AttrBuilder.contains('StrArrayAttr')) else
      _ods_ir.AttrBuilder.get('StrArrayAttr')(names, context=_ods_context))
    results = []
    _ods_successors = None
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

  @builtins.property
  def args(self) -> _ods_ir.OpOperandList:
    _ods_variadic_group_length = len(self.operation.operands) - 1 + 1
    return self.operation.operands[0:0 + _ods_variadic_group_length]

  @builtins.property
  def names(self) -> _ods_ir.ArrayAttr:
    return self.operation.attributes["names"]

  @names.setter
  def names(self, value: _ods_ir.ArrayAttr):
    if value is None:
      raise ValueError("'None' not allowed as value for mandatory attributes")
    self.operation.attributes["names"] = value

@_ods_cext.register_op_adaptor(RegionsOp)
class RegionsOpAdaptor(_ods_ir.OpAdaptor):
  OPERATION_NAME = "irdl.regions"

  @builtins.property
  def args(self) -> _ods_ir.OpOperandList:
    _ods_variadic_group_length = len(self.operands) - 1 + 1
    return self.operands[0:0 + _ods_variadic_group_length]

  @builtins.property
  def names(self) -> _ods_ir.ArrayAttr:
    return self.attributes["names"]

def regions_(args: _Sequence[_ods_ir.Value], names: _Union[_Sequence[str], _ods_ir.ArrayAttr], *, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None) -> RegionsOp:
  return RegionsOp(args=args, names=names, loc=loc, ip=ip)

@_ods_cext.register_operation(_Dialect)
class ResultsOp(_ods_ir.OpView):
  r"""
  `irdl.results` define the results of the `irdl.operation` parent operation
  definition. Each result is named after an identifier.
  
  In the following example, `irdl.results` defines the results of the
  `get_values` operation:
  
  ```mlir
  irdl.dialect @cmath {
  
    irdl.type @complex { /* ... */ }
  
    /// Returns the real and imaginary parts of a complex number.
    irdl.operation @get_values {
      %0 = irdl.any
      %1 = irdl.parametric @cmath::@complex<%0>
      irdl.results(re: %0, im: %0)
      irdl.operands(complex: %1)
    }
  }
  ```
  
  The operation will expect one operand of the `cmath.complex` type, and two
  results that have the underlying type of the `cmath.complex`.
  
  The results can also be marked as variadic or optional:
  ```mlir
  irdl.results(foo: %0, bar: single %1, baz: optional %2, qux: variadic %3)
  ```
  
  Here, foo and bar are required single results, baz is an optional result,
  and qux is a variadic result.
  
  When more than one result is marked as optional or variadic, the operation
  will expect a 'resultSegmentSizes' attribute that defines the number of
  results in each segment.
  """

  OPERATION_NAME = "irdl.results"

  _ODS_REGIONS = (0, True)

  def __init__(self, args: _Sequence[_ods_ir.Value], names: _Union[_Sequence[str], _ods_ir.ArrayAttr], variadicity: _Union[_Any, _ods_ir.Attribute], *, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None):
    operands = []
    attributes = {}
    regions = None
    operands.extend(_get_op_results_or_values(args))
    _ods_context = _ods_get_default_loc_context(loc)
    attributes["names"] = (names if (
    isinstance(names, _ods_ir.Attribute) or
    not _ods_ir.AttrBuilder.contains('StrArrayAttr')) else
      _ods_ir.AttrBuilder.get('StrArrayAttr')(names, context=_ods_context))
    attributes["variadicity"] = (variadicity if (
    isinstance(variadicity, _ods_ir.Attribute) or
    not _ods_ir.AttrBuilder.contains('VariadicityArrayAttr')) else
      _ods_ir.AttrBuilder.get('VariadicityArrayAttr')(variadicity, context=_ods_context))
    results = []
    _ods_successors = None
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

  @builtins.property
  def args(self) -> _ods_ir.OpOperandList:
    _ods_variadic_group_length = len(self.operation.operands) - 1 + 1
    return self.operation.operands[0:0 + _ods_variadic_group_length]

  @builtins.property
  def names(self) -> _ods_ir.ArrayAttr:
    return self.operation.attributes["names"]

  @names.setter
  def names(self, value: _ods_ir.ArrayAttr):
    if value is None:
      raise ValueError("'None' not allowed as value for mandatory attributes")
    self.operation.attributes["names"] = value

  @builtins.property
  def variadicity(self) -> _ods_ir.Attribute:
    return self.operation.attributes["variadicity"]

  @variadicity.setter
  def variadicity(self, value: _ods_ir.Attribute):
    if value is None:
      raise ValueError("'None' not allowed as value for mandatory attributes")
    self.operation.attributes["variadicity"] = value

@_ods_cext.register_op_adaptor(ResultsOp)
class ResultsOpAdaptor(_ods_ir.OpAdaptor):
  OPERATION_NAME = "irdl.results"

  @builtins.property
  def args(self) -> _ods_ir.OpOperandList:
    _ods_variadic_group_length = len(self.operands) - 1 + 1
    return self.operands[0:0 + _ods_variadic_group_length]

  @builtins.property
  def names(self) -> _ods_ir.ArrayAttr:
    return self.attributes["names"]

  @builtins.property
  def variadicity(self) -> _ods_ir.Attribute:
    return self.attributes["variadicity"]

def results_(args: _Sequence[_ods_ir.Value], names: _Union[_Sequence[str], _ods_ir.ArrayAttr], variadicity: _Union[_Any, _ods_ir.Attribute], *, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None) -> ResultsOp:
  return ResultsOp(args=args, names=names, variadicity=variadicity, loc=loc, ip=ip)

@_ods_cext.register_operation(_Dialect)
class TypeOp(_ods_ir.OpView):
  r"""
  `irdl.type` defines a new type belonging to the `irdl.dialect` parent.
  
  The type parameters can be defined with an `irdl.parameters` operation in
  the optional region.
  
  Example:
  
  ```mlir
  irdl.dialect @cmath {
    irdl.type @complex {
      %0 = irdl.is i32
      %1 = irdl.is i64
      %2 = irdl.any_of(%0, %1)
      irdl.parameters(%2)
    }
  }
  ```
  
  The above program defines a type `complex` inside the dialect `cmath`. The
  type has a single parameter that should be either `i32` or `i64`.
  """

  OPERATION_NAME = "irdl.type"

  _ODS_REGIONS = (1, True)

  def __init__(self, sym_name: _Union[str, _ods_ir.StringAttr], *, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None):
    operands = []
    attributes = {}
    regions = None
    _ods_context = _ods_get_default_loc_context(loc)
    attributes["sym_name"] = (sym_name if (
    isinstance(sym_name, _ods_ir.Attribute) or
    not _ods_ir.AttrBuilder.contains('SymbolNameAttr')) else
      _ods_ir.AttrBuilder.get('SymbolNameAttr')(sym_name, context=_ods_context))
    results = []
    _ods_successors = None
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

  @builtins.property
  def sym_name(self) -> _ods_ir.StringAttr:
    return self.operation.attributes["sym_name"]

  @sym_name.setter
  def sym_name(self, value: _ods_ir.StringAttr):
    if value is None:
      raise ValueError("'None' not allowed as value for mandatory attributes")
    self.operation.attributes["sym_name"] = value

  @builtins.property
  def body(self) -> _ods_ir.Region:
    return self.regions[0]

@_ods_cext.register_op_adaptor(TypeOp)
class TypeOpAdaptor(_ods_ir.OpAdaptor):
  OPERATION_NAME = "irdl.type"

  @builtins.property
  def sym_name(self) -> _ods_ir.StringAttr:
    return self.attributes["sym_name"]

def type_(sym_name: _Union[str, _ods_ir.StringAttr], *, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None) -> TypeOp:
  return TypeOp(sym_name=sym_name, loc=loc, ip=ip)



# SPDX-FileCopyrightText: Copyright (c) 2026 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

"""
NUMBA_CUDA_MLIR Universal Function (ufunc) Database

This module provides a database mapping NumPy ufuncs to their MLIR lowering
implementations. When NumPy ufuncs are called on arrays, Numba routes them through
the 'arrayexpr' code path, which requires special handling to map the ufunc to
the appropriate MLIR code generator.

The database is lazily initialized to avoid circular imports and provides a simple
lookup mechanism for the arrayexpr lowering process.
"""

import numpy as np
import operator
from typing import Callable, Dict, Any, Optional

# Lazily initialized ufunc database
_ufunc_db: Optional[Dict[Any, Callable]] = None


def _lazy_init_db():
    """Initialize the ufunc database on first access to avoid circular imports."""
    global _ufunc_db

    if _ufunc_db is None:
        from numba_cuda_mlir.lowering import (
            numpy,
            builtins,
        )

        _ufunc_db = {}
        _ufunc_db.update(numpy.ufunc_registry.registry)
        _ufunc_db.update(builtins.ufunc_registry.registry)


def get_ufunc_lowering(ufunc) -> Optional[Callable]:
    """
    Get the MLIR lowering function for a given NumPy ufunc.

    Args:
        ufunc: A NumPy universal function object (e.g., np.abs, np.sqrt)

    Returns:
        The MLIR code generation function if the ufunc is supported, None otherwise.
        The returned function has the signature:
        (builder, target, args, kwargs) -> None

    Example:
        >>> lowering = get_ufunc_lowering(np.abs)
        >>> if lowering:
        >>>     lowering(builder, target, args, {})
    """
    _lazy_init_db()
    assert _ufunc_db is not None
    return _ufunc_db.get(ufunc)


def is_supported_ufunc(ufunc) -> bool:
    """
    Check if a NumPy ufunc is supported for MLIR lowering.

    Args:
        ufunc: A NumPy universal function object

    Returns:
        True if the ufunc has a registered MLIR lowering, False otherwise.
    """
    _lazy_init_db()
    assert _ufunc_db is not None
    return ufunc in _ufunc_db


def get_supported_ufuncs():
    """
    Get a list of all supported NumPy ufuncs.

    Returns:
        A list of NumPy ufunc objects that have MLIR lowerings registered.
    """
    _lazy_init_db()
    assert _ufunc_db is not None
    return list(_ufunc_db.keys())




# Autogenerated by mlir-tblgen; don't manually edit.

from enum import IntEnum, auto, IntFlag
from ._ods_common import _cext as _ods_cext
from ..ir import register_attribute_builder
_ods_ir = _ods_cext.ir

class AsmDialect(IntEnum):
    """ATT (0) or Intel (1) asm dialect"""

    AD_ATT = 0
    AD_Intel = 1

    def __str__(self):
        if self is AsmDialect.AD_ATT:
            return "att"
        if self is AsmDialect.AD_Intel:
            return "intel"
        raise ValueError("Unknown AsmDialect enum entry.")



@register_attribute_builder("AsmATTOrIntel", allow_existing=True)
def _asmattorintel(x, context):
    return _ods_ir.IntegerAttr.get(_ods_ir.IntegerType.get_signless(64, context=context), int(x))

class AtomicBinOp(IntEnum):
    """llvm.atomicrmw binary operations"""

    xchg = 0
    add = 1
    sub = 2
    _and = 3
    nand = 4
    _or = 5
    _xor = 6
    max = 7
    min = 8
    umax = 9
    umin = 10
    fadd = 11
    fsub = 12
    fmax = 13
    fmin = 14
    uinc_wrap = 15
    udec_wrap = 16
    usub_cond = 17
    usub_sat = 18
    fmaximum = 19
    fminimum = 20
    fmaximumnum = 21
    fminimumnum = 22

    def __str__(self):
        if self is AtomicBinOp.xchg:
            return "xchg"
        if self is AtomicBinOp.add:
            return "add"
        if self is AtomicBinOp.sub:
            return "sub"
        if self is AtomicBinOp._and:
            return "_and"
        if self is AtomicBinOp.nand:
            return "nand"
        if self is AtomicBinOp._or:
            return "_or"
        if self is AtomicBinOp._xor:
            return "_xor"
        if self is AtomicBinOp.max:
            return "max"
        if self is AtomicBinOp.min:
            return "min"
        if self is AtomicBinOp.umax:
            return "umax"
        if self is AtomicBinOp.umin:
            return "umin"
        if self is AtomicBinOp.fadd:
            return "fadd"
        if self is AtomicBinOp.fsub:
            return "fsub"
        if self is AtomicBinOp.fmax:
            return "fmax"
        if self is AtomicBinOp.fmin:
            return "fmin"
        if self is AtomicBinOp.uinc_wrap:
            return "uinc_wrap"
        if self is AtomicBinOp.udec_wrap:
            return "udec_wrap"
        if self is AtomicBinOp.usub_cond:
            return "usub_cond"
        if self is AtomicBinOp.usub_sat:
            return "usub_sat"
        if self is AtomicBinOp.fmaximum:
            return "fmaximum"
        if self is AtomicBinOp.fminimum:
            return "fminimum"
        if self is AtomicBinOp.fmaximumnum:
            return "fmaximumnum"
        if self is AtomicBinOp.fminimumnum:
            return "fminimumnum"
        raise ValueError("Unknown AtomicBinOp enum entry.")



@register_attribute_builder("AtomicBinOp", allow_existing=True)
def _atomicbinop(x, context):
    return _ods_ir.IntegerAttr.get(_ods_ir.IntegerType.get_signless(64, context=context), int(x))

class AtomicOrdering(IntEnum):
    """Atomic ordering for LLVM's memory model"""

    not_atomic = 0
    unordered = 1
    monotonic = 2
    acquire = 4
    release = 5
    acq_rel = 6
    seq_cst = 7

    def __str__(self):
        if self is AtomicOrdering.not_atomic:
            return "not_atomic"
        if self is AtomicOrdering.unordered:
            return "unordered"
        if self is AtomicOrdering.monotonic:
            return "monotonic"
        if self is AtomicOrdering.acquire:
            return "acquire"
        if self is AtomicOrdering.release:
            return "release"
        if self is AtomicOrdering.acq_rel:
            return "acq_rel"
        if self is AtomicOrdering.seq_cst:
            return "seq_cst"
        raise ValueError("Unknown AtomicOrdering enum entry.")



@register_attribute_builder("AtomicOrdering", allow_existing=True)
def _atomicordering(x, context):
    return _ods_ir.IntegerAttr.get(_ods_ir.IntegerType.get_signless(64, context=context), int(x))

class CConv(IntEnum):
    """Calling Conventions"""

    C = 0
    Fast = 8
    Cold = 9
    GHC = 10
    HiPE = 11
    AnyReg = 13
    PreserveMost = 14
    PreserveAll = 15
    Swift = 16
    CXX_FAST_TLS = 17
    Tail = 18
    CFGuard_Check = 19
    SwiftTail = 20
    X86_StdCall = 64
    X86_FastCall = 65
    ARM_APCS = 66
    ARM_AAPCS = 67
    ARM_AAPCS_VFP = 68
    MSP430_INTR = 69
    X86_ThisCall = 70
    PTX_Kernel = 71
    PTX_Device = 72
    SPIR_FUNC = 75
    SPIR_KERNEL = 76
    Intel_OCL_BI = 77
    X86_64_SysV = 78
    Win64 = 79
    X86_VectorCall = 80
    DUMMY_HHVM = 81
    DUMMY_HHVM_C = 82
    X86_INTR = 83
    AVR_INTR = 84
    AVR_BUILTIN = 86
    AMDGPU_VS = 87
    AMDGPU_GS = 88
    AMDGPU_CS = 90
    AMDGPU_KERNEL = 91
    X86_RegCall = 92
    AMDGPU_HS = 93
    MSP430_BUILTIN = 94
    AMDGPU_LS = 95
    AMDGPU_ES = 96
    AArch64_VectorCall = 97
    AArch64_SVE_VectorCall = 98
    WASM_EmscriptenInvoke = 99
    AMDGPU_Gfx = 100
    M68k_INTR = 101

    def __str__(self):
        if self is CConv.C:
            return "ccc"
        if self is CConv.Fast:
            return "fastcc"
        if self is CConv.Cold:
            return "coldcc"
        if self is CConv.GHC:
            return "cc_10"
        if self is CConv.HiPE:
            return "cc_11"
        if self is CConv.AnyReg:
            return "anyregcc"
        if self is CConv.PreserveMost:
            return "preserve_mostcc"
        if self is CConv.PreserveAll:
            return "preserve_allcc"
        if self is CConv.Swift:
            return "swiftcc"
        if self is CConv.CXX_FAST_TLS:
            return "cxx_fast_tlscc"
        if self is CConv.Tail:
            return "tailcc"
        if self is CConv.CFGuard_Check:
            return "cfguard_checkcc"
        if self is CConv.SwiftTail:
            return "swifttailcc"
        if self is CConv.X86_StdCall:
            return "x86_stdcallcc"
        if self is CConv.X86_FastCall:
            return "x86_fastcallcc"
        if self is CConv.ARM_APCS:
            return "arm_apcscc"
        if self is CConv.ARM_AAPCS:
            return "arm_aapcscc"
        if self is CConv.ARM_AAPCS_VFP:
            return "arm_aapcs_vfpcc"
        if self is CConv.MSP430_INTR:
            return "msp430_intrcc"
        if self is CConv.X86_ThisCall:
            return "x86_thiscallcc"
        if self is CConv.PTX_Kernel:
            return "ptx_kernelcc"
        if self is CConv.PTX_Device:
            return "ptx_devicecc"
        if self is CConv.SPIR_FUNC:
            return "spir_funccc"
        if self is CConv.SPIR_KERNEL:
            return "spir_kernelcc"
        if self is CConv.Intel_OCL_BI:
            return "intel_ocl_bicc"
        if self is CConv.X86_64_SysV:
            return "x86_64_sysvcc"
        if self is CConv.Win64:
            return "win64cc"
        if self is CConv.X86_VectorCall:
            return "x86_vectorcallcc"
        if self is CConv.DUMMY_HHVM:
            return "hhvmcc"
        if self is CConv.DUMMY_HHVM_C:
            return "hhvm_ccc"
        if self is CConv.X86_INTR:
            return "x86_intrcc"
        if self is CConv.AVR_INTR:
            return "avr_intrcc"
        if self is CConv.AVR_BUILTIN:
            return "avr_builtincc"
        if self is CConv.AMDGPU_VS:
            return "amdgpu_vscc"
        if self is CConv.AMDGPU_GS:
            return "amdgpu_gscc"
        if self is CConv.AMDGPU_CS:
            return "amdgpu_cscc"
        if self is CConv.AMDGPU_KERNEL:
            return "amdgpu_kernelcc"
        if self is CConv.X86_RegCall:
            return "x86_regcallcc"
        if self is CConv.AMDGPU_HS:
            return "amdgpu_hscc"
        if self is CConv.MSP430_BUILTIN:
            return "msp430_builtincc"
        if self is CConv.AMDGPU_LS:
            return "amdgpu_lscc"
        if self is CConv.AMDGPU_ES:
            return "amdgpu_escc"
        if self is CConv.AArch64_VectorCall:
            return "aarch64_vectorcallcc"
        if self is CConv.AArch64_SVE_VectorCall:
            return "aarch64_sve_vectorcallcc"
        if self is CConv.WASM_EmscriptenInvoke:
            return "wasm_emscripten_invokecc"
        if self is CConv.AMDGPU_Gfx:
            return "amdgpu_gfxcc"
        if self is CConv.M68k_INTR:
            return "m68k_intrcc"
        raise ValueError("Unknown CConv enum entry.")



@register_attribute_builder("CConvEnum", allow_existing=True)
def _cconvenum(x, context):
    return _ods_ir.IntegerAttr.get(_ods_ir.IntegerType.get_signless(64, context=context), int(x))

class Comdat(IntEnum):
    """LLVM Comdat Types"""

    Any = 0
    ExactMatch = 1
    Largest = 2
    NoDeduplicate = 3
    SameSize = 4

    def __str__(self):
        if self is Comdat.Any:
            return "any"
        if self is Comdat.ExactMatch:
            return "exactmatch"
        if self is Comdat.Largest:
            return "largest"
        if self is Comdat.NoDeduplicate:
            return "nodeduplicate"
        if self is Comdat.SameSize:
            return "samesize"
        raise ValueError("Unknown Comdat enum entry.")



@register_attribute_builder("Comdat", allow_existing=True)
def _comdat(x, context):
    return _ods_ir.IntegerAttr.get(_ods_ir.IntegerType.get_signless(64, context=context), int(x))

class DIFlags(IntFlag):
    """LLVM DI flags"""

    Zero = 0
    Bit0 = 1
    Bit1 = 2
    Private = 1
    Protected = 2
    Public = 3
    FwdDecl = 4
    AppleBlock = 8
    ReservedBit4 = 16
    Virtual = 32
    Artificial = 64
    Explicit = 128
    Prototyped = 256
    ObjcClassComplete = 512
    ObjectPointer = 1024
    Vector = 2048
    StaticMember = 4096
    LValueReference = 8192
    RValueReference = 16384
    ExportSymbols = 32768
    SingleInheritance = 65536
    MultipleInheritance = 65536
    VirtualInheritance = 65536
    IntroducedVirtual = 262144
    BitField = 524288
    NoReturn = 1048576
    TypePassByValue = 4194304
    TypePassByReference = 8388608
    EnumClass = 16777216
    Thunk = 33554432
    NonTrivial = 67108864
    BigEndian = 134217728
    LittleEndian = 268435456
    AllCallsDescribed = 536870912

    def __iter__(self):
        return iter([case for case in type(self) if (self & case) is case and self is not case])
    def __len__(self):
        return bin(self).count("1")

    def __str__(self):
        if len(self) > 1:
            return "|".join(map(str, self))
        if self is DIFlags.Zero:
            return "Zero"
        if self is DIFlags.Bit0:
            return "Bit0"
        if self is DIFlags.Bit1:
            return "Bit1"
        if self is DIFlags.Private:
            return "Private"
        if self is DIFlags.Protected:
            return "Protected"
        if self is DIFlags.Public:
            return "Public"
        if self is DIFlags.FwdDecl:
            return "FwdDecl"
        if self is DIFlags.AppleBlock:
            return "AppleBlock"
        if self is DIFlags.ReservedBit4:
            return "ReservedBit4"
        if self is DIFlags.Virtual:
            return "Virtual"
        if self is DIFlags.Artificial:
            return "Artificial"
        if self is DIFlags.Explicit:
            return "Explicit"
        if self is DIFlags.Prototyped:
            return "Prototyped"
        if self is DIFlags.ObjcClassComplete:
            return "ObjcClassComplete"
        if self is DIFlags.ObjectPointer:
            return "ObjectPointer"
        if self is DIFlags.Vector:
            return "Vector"
        if self is DIFlags.StaticMember:
            return "StaticMember"
        if self is DIFlags.LValueReference:
            return "LValueReference"
        if self is DIFlags.RValueReference:
            return "RValueReference"
        if self is DIFlags.ExportSymbols:
            return "ExportSymbols"
        if self is DIFlags.SingleInheritance:
            return "SingleInheritance"
        if self is DIFlags.MultipleInheritance:
            return "MultipleInheritance"
        if self is DIFlags.VirtualInheritance:
            return "VirtualInheritance"
        if self is DIFlags.IntroducedVirtual:
            return "IntroducedVirtual"
        if self is DIFlags.BitField:
            return "BitField"
        if self is DIFlags.NoReturn:
            return "NoReturn"
        if self is DIFlags.TypePassByValue:
            return "TypePassByValue"
        if self is DIFlags.TypePassByReference:
            return "TypePassByReference"
        if self is DIFlags.EnumClass:
            return "EnumClass"
        if self is DIFlags.Thunk:
            return "Thunk"
        if self is DIFlags.NonTrivial:
            return "NonTrivial"
        if self is DIFlags.BigEndian:
            return "BigEndian"
        if self is DIFlags.LittleEndian:
            return "LittleEndian"
        if self is DIFlags.AllCallsDescribed:
            return "AllCallsDescribed"
        raise ValueError("Unknown DIFlags enum entry.")



@register_attribute_builder("DIFlags", allow_existing=True)
def _diflags(x, context):
    return _ods_ir.IntegerAttr.get(_ods_ir.IntegerType.get_signless(32, context=context), int(x))

class DISubprogramFlags(IntFlag):
    """LLVM DISubprogram flags"""

    Virtual = 1
    PureVirtual = 2
    LocalToUnit = 4
    Definition = 8
    Optimized = 16
    Pure = 32
    Elemental = 64
    Recursive = 128
    MainSubprogram = 256
    Deleted = 512
    ObjCDirect = 2048

    def __iter__(self):
        return iter([case for case in type(self) if (self & case) is case and self is not case])
    def __len__(self):
        return bin(self).count("1")

    def __str__(self):
        if len(self) > 1:
            return "|".join(map(str, self))
        if self is DISubprogramFlags.Virtual:
            return "Virtual"
        if self is DISubprogramFlags.PureVirtual:
            return "PureVirtual"
        if self is DISubprogramFlags.LocalToUnit:
            return "LocalToUnit"
        if self is DISubprogramFlags.Definition:
            return "Definition"
        if self is DISubprogramFlags.Optimized:
            return "Optimized"
        if self is DISubprogramFlags.Pure:
            return "Pure"
        if self is DISubprogramFlags.Elemental:
            return "Elemental"
        if self is DISubprogramFlags.Recursive:
            return "Recursive"
        if self is DISubprogramFlags.MainSubprogram:
            return "MainSubprogram"
        if self is DISubprogramFlags.Deleted:
            return "Deleted"
        if self is DISubprogramFlags.ObjCDirect:
            return "ObjCDirect"
        raise ValueError("Unknown DISubprogramFlags enum entry.")



@register_attribute_builder("DISubprogramFlags", allow_existing=True)
def _disubprogramflags(x, context):
    return _ods_ir.IntegerAttr.get(_ods_ir.IntegerType.get_signless(32, context=context), int(x))

class DenormalModeKind(IntEnum):
    """LLVM Denormal Mode Kinds"""

    IEEE = 0
    PreserveSign = 1
    PositiveZero = 2
    Dynamic = 3
    Invalid = auto()

    def __str__(self):
        if self is DenormalModeKind.IEEE:
            return "ieee"
        if self is DenormalModeKind.PreserveSign:
            return "preservesign"
        if self is DenormalModeKind.PositiveZero:
            return "positivezero"
        if self is DenormalModeKind.Dynamic:
            return "dynamic"
        if self is DenormalModeKind.Invalid:
            return "invalid"
        raise ValueError("Unknown DenormalModeKind enum entry.")



@register_attribute_builder("DenormalModeKindAttr", allow_existing=True)
def _denormalmodekindattr(x, context):
    return _ods_ir.IntegerAttr.get(_ods_ir.IntegerType.get_signless(64, context=context), int(x))

class FCmpPredicate(IntEnum):
    """llvm.fcmp comparison predicate"""

    _false = 0
    oeq = 1
    ogt = 2
    oge = 3
    olt = 4
    ole = 5
    one = 6
    ord = 7
    ueq = 8
    ugt = 9
    uge = 10
    ult = 11
    ule = 12
    une = 13
    uno = 14
    _true = 15

    def __str__(self):
        if self is FCmpPredicate._false:
            return "_false"
        if self is FCmpPredicate.oeq:
            return "oeq"
        if self is FCmpPredicate.ogt:
            return "ogt"
        if self is FCmpPredicate.oge:
            return "oge"
        if self is FCmpPredicate.olt:
            return "olt"
        if self is FCmpPredicate.ole:
            return "ole"
        if self is FCmpPredicate.one:
            return "one"
        if self is FCmpPredicate.ord:
            return "ord"
        if self is FCmpPredicate.ueq:
            return "ueq"
        if self is FCmpPredicate.ugt:
            return "ugt"
        if self is FCmpPredicate.uge:
            return "uge"
        if self is FCmpPredicate.ult:
            return "ult"
        if self is FCmpPredicate.ule:
            return "ule"
        if self is FCmpPredicate.une:
            return "une"
        if self is FCmpPredicate.uno:
            return "uno"
        if self is FCmpPredicate._true:
            return "_true"
        raise ValueError("Unknown FCmpPredicate enum entry.")



@register_attribute_builder("FCmpPredicate", allow_existing=True)
def _fcmppredicate(x, context):
    return _ods_ir.IntegerAttr.get(_ods_ir.IntegerType.get_signless(64, context=context), int(x))

class FPExceptionBehavior(IntEnum):
    """LLVM Exception Behavior"""

    Ignore = 0
    MayTrap = 1
    Strict = 2

    def __str__(self):
        if self is FPExceptionBehavior.Ignore:
            return "ignore"
        if self is FPExceptionBehavior.MayTrap:
            return "maytrap"
        if self is FPExceptionBehavior.Strict:
            return "strict"
        raise ValueError("Unknown FPExceptionBehavior enum entry.")



@register_attribute_builder("FPExceptionBehaviorAttr", allow_existing=True)
def _fpexceptionbehaviorattr(x, context):
    return _ods_ir.IntegerAttr.get(_ods_ir.IntegerType.get_signless(64, context=context), int(x))

class FastmathFlags(IntFlag):
    """LLVM fastmath flags"""

    none = 0
    nnan = 1
    ninf = 2
    nsz = 4
    arcp = 8
    contract = 16
    afn = 32
    reassoc = 64
    fast = 127

    def __iter__(self):
        return iter([case for case in type(self) if (self & case) is case and self is not case])
    def __len__(self):
        return bin(self).count("1")

    def __str__(self):
        if len(self) > 1:
            return ", ".join(map(str, self))
        if self is FastmathFlags.none:
            return "none"
        if self is FastmathFlags.nnan:
            return "nnan"
        if self is FastmathFlags.ninf:
            return "ninf"
        if self is FastmathFlags.nsz:
            return "nsz"
        if self is FastmathFlags.arcp:
            return "arcp"
        if self is FastmathFlags.contract:
            return "contract"
        if self is FastmathFlags.afn:
            return "afn"
        if self is FastmathFlags.reassoc:
            return "reassoc"
        if self is FastmathFlags.fast:
            return "fast"
        raise ValueError("Unknown FastmathFlags enum entry.")



@register_attribute_builder("FastmathFlags", allow_existing=True)
def _fastmathflags(x, context):
    return _ods_ir.IntegerAttr.get(_ods_ir.IntegerType.get_signless(32, context=context), int(x))

class FramePointerKind(IntEnum):
    """LLVM FramePointerKind"""

    None_ = 0
    NonLeaf = 1
    All = 2
    Reserved = 3
    NonLeafNoReserve = 4

    def __str__(self):
        if self is FramePointerKind.None_:
            return "none"
        if self is FramePointerKind.NonLeaf:
            return "non-leaf"
        if self is FramePointerKind.All:
            return "all"
        if self is FramePointerKind.Reserved:
            return "reserved"
        if self is FramePointerKind.NonLeafNoReserve:
            return "non-leaf-no-reserve"
        raise ValueError("Unknown FramePointerKind enum entry.")



@register_attribute_builder("FramePointerKindEnum", allow_existing=True)
def _framepointerkindenum(x, context):
    return _ods_ir.IntegerAttr.get(_ods_ir.IntegerType.get_signless(64, context=context), int(x))

class GEPNoWrapFlags(IntFlag):
    """::mlir::LLVM::GEPNoWrapFlags"""

    none = 0
    inboundsFlag = 1
    nusw = 2
    nuw = 4
    inbounds = 3

    def __iter__(self):
        return iter([case for case in type(self) if (self & case) is case and self is not case])
    def __len__(self):
        return bin(self).count("1")

    def __str__(self):
        if len(self) > 1:
            return "|".join(map(str, self))
        if self is GEPNoWrapFlags.none:
            return "none"
        if self is GEPNoWrapFlags.inboundsFlag:
            return "inbounds_flag"
        if self is GEPNoWrapFlags.nusw:
            return "nusw"
        if self is GEPNoWrapFlags.nuw:
            return "nuw"
        if self is GEPNoWrapFlags.inbounds:
            return "inbounds"
        raise ValueError("Unknown GEPNoWrapFlags enum entry.")



class ICmpPredicate(IntEnum):
    """llvm.icmp comparison predicate"""

    eq = 0
    ne = 1
    slt = 2
    sle = 3
    sgt = 4
    sge = 5
    ult = 6
    ule = 7
    ugt = 8
    uge = 9

    def __str__(self):
        if self is ICmpPredicate.eq:
            return "eq"
        if self is ICmpPredicate.ne:
            return "ne"
        if self is ICmpPredicate.slt:
            return "slt"
        if self is ICmpPredicate.sle:
            return "sle"
        if self is ICmpPredicate.sgt:
            return "sgt"
        if self is ICmpPredicate.sge:
            return "sge"
        if self is ICmpPredicate.ult:
            return "ult"
        if self is ICmpPredicate.ule:
            return "ule"
        if self is ICmpPredicate.ugt:
            return "ugt"
        if self is ICmpPredicate.uge:
            return "uge"
        raise ValueError("Unknown ICmpPredicate enum entry.")



@register_attribute_builder("ICmpPredicate", allow_existing=True)
def _icmppredicate(x, context):
    return _ods_ir.IntegerAttr.get(_ods_ir.IntegerType.get_signless(64, context=context), int(x))

class IntegerOverflowFlags(IntFlag):
    """LLVM integer overflow flags"""

    none = 0
    nsw = 1
    nuw = 2

    def __iter__(self):
        return iter([case for case in type(self) if (self & case) is case and self is not case])
    def __len__(self):
        return bin(self).count("1")

    def __str__(self):
        if len(self) > 1:
            return ", ".join(map(str, self))
        if self is IntegerOverflowFlags.none:
            return "none"
        if self is IntegerOverflowFlags.nsw:
            return "nsw"
        if self is IntegerOverflowFlags.nuw:
            return "nuw"
        raise ValueError("Unknown IntegerOverflowFlags enum entry.")



class DIEmissionKind(IntEnum):
    """LLVM debug emission kind"""

    None_ = 0
    Full = 1
    LineTablesOnly = 2
    DebugDirectivesOnly = 3

    def __str__(self):
        if self is DIEmissionKind.None_:
            return "None"
        if self is DIEmissionKind.Full:
            return "Full"
        if self is DIEmissionKind.LineTablesOnly:
            return "LineTablesOnly"
        if self is DIEmissionKind.DebugDirectivesOnly:
            return "DebugDirectivesOnly"
        raise ValueError("Unknown DIEmissionKind enum entry.")



@register_attribute_builder("LLVM_DIEmissionKind", allow_existing=True)
def _llvm_diemissionkind(x, context):
    return _ods_ir.IntegerAttr.get(_ods_ir.IntegerType.get_signless(64, context=context), int(x))

class DINameTableKind(IntEnum):
    """LLVM debug name table kind"""

    Default = 0
    GNU = 1
    None_ = 2
    Apple = 3

    def __str__(self):
        if self is DINameTableKind.Default:
            return "Default"
        if self is DINameTableKind.GNU:
            return "GNU"
        if self is DINameTableKind.None_:
            return "None"
        if self is DINameTableKind.Apple:
            return "Apple"
        raise ValueError("Unknown DINameTableKind enum entry.")



@register_attribute_builder("LLVM_DINameTableKind", allow_existing=True)
def _llvm_dinametablekind(x, context):
    return _ods_ir.IntegerAttr.get(_ods_ir.IntegerType.get_signless(64, context=context), int(x))

class ProfileSummaryFormatKind(IntEnum):
    """LLVM ProfileSummary format kinds"""

    SampleProfile = 0
    InstrProf = 1
    CSInstrProf = 2

    def __str__(self):
        if self is ProfileSummaryFormatKind.SampleProfile:
            return "SampleProfile"
        if self is ProfileSummaryFormatKind.InstrProf:
            return "InstrProf"
        if self is ProfileSummaryFormatKind.CSInstrProf:
            return "CSInstrProf"
        raise ValueError("Unknown ProfileSummaryFormatKind enum entry.")



@register_attribute_builder("LLVM_ProfileSummaryFormatKind", allow_existing=True)
def _llvm_profilesummaryformatkind(x, context):
    return _ods_ir.IntegerAttr.get(_ods_ir.IntegerType.get_signless(64, context=context), int(x))

class Linkage(IntEnum):
    """LLVM linkage types"""

    External = 0
    AvailableExternally = 1
    Linkonce = 2
    LinkonceODR = 3
    Weak = 4
    WeakODR = 5
    Appending = 6
    Internal = 7
    Private = 8
    ExternWeak = 9
    Common = 10

    def __str__(self):
        if self is Linkage.External:
            return "external"
        if self is Linkage.AvailableExternally:
            return "available_externally"
        if self is Linkage.Linkonce:
            return "linkonce"
        if self is Linkage.LinkonceODR:
            return "linkonce_odr"
        if self is Linkage.Weak:
            return "weak"
        if self is Linkage.WeakODR:
            return "weak_odr"
        if self is Linkage.Appending:
            return "appending"
        if self is Linkage.Internal:
            return "internal"
        if self is Linkage.Private:
            return "private"
        if self is Linkage.ExternWeak:
            return "extern_weak"
        if self is Linkage.Common:
            return "common"
        raise ValueError("Unknown Linkage enum entry.")



@register_attribute_builder("LinkageEnum", allow_existing=True)
def _linkageenum(x, context):
    return _ods_ir.IntegerAttr.get(_ods_ir.IntegerType.get_signless(64, context=context), int(x))

class ModFlagBehavior(IntEnum):
    """LLVM Module Flag Behavior"""

    Error = 1
    Warning = 2
    Require = 3
    Override = 4
    Append = 5
    AppendUnique = 6
    Max = 7
    Min = 8

    def __str__(self):
        if self is ModFlagBehavior.Error:
            return "error"
        if self is ModFlagBehavior.Warning:
            return "warning"
        if self is ModFlagBehavior.Require:
            return "require"
        if self is ModFlagBehavior.Override:
            return "override"
        if self is ModFlagBehavior.Append:
            return "append"
        if self is ModFlagBehavior.AppendUnique:
            return "append_unique"
        if self is ModFlagBehavior.Max:
            return "max"
        if self is ModFlagBehavior.Min:
            return "min"
        raise ValueError("Unknown ModFlagBehavior enum entry.")



@register_attribute_builder("ModFlagBehaviorAttr", allow_existing=True)
def _modflagbehaviorattr(x, context):
    return _ods_ir.IntegerAttr.get(_ods_ir.IntegerType.get_signless(64, context=context), int(x))

class ModRefInfo(IntEnum):
    """LLVM ModRefInfo"""

    NoModRef = 0
    Ref = 1
    Mod = 2
    ModRef = 3

    def __str__(self):
        if self is ModRefInfo.NoModRef:
            return "none"
        if self is ModRefInfo.Ref:
            return "read"
        if self is ModRefInfo.Mod:
            return "write"
        if self is ModRefInfo.ModRef:
            return "readwrite"
        raise ValueError("Unknown ModRefInfo enum entry.")



@register_attribute_builder("ModRefInfoEnum", allow_existing=True)
def _modrefinfoenum(x, context):
    return _ods_ir.IntegerAttr.get(_ods_ir.IntegerType.get_signless(64, context=context), int(x))

class RoundingMode(IntEnum):
    """LLVM Rounding Mode"""

    TowardZero = 0
    NearestTiesToEven = 1
    TowardPositive = 2
    TowardNegative = 3
    NearestTiesToAway = 4
    Dynamic = 7
    Invalid = auto()

    def __str__(self):
        if self is RoundingMode.TowardZero:
            return "towardzero"
        if self is RoundingMode.NearestTiesToEven:
            return "tonearest"
        if self is RoundingMode.TowardPositive:
            return "upward"
        if self is RoundingMode.TowardNegative:
            return "downward"
        if self is RoundingMode.NearestTiesToAway:
            return "tonearestaway"
        if self is RoundingMode.Dynamic:
            return "dynamic"
        if self is RoundingMode.Invalid:
            return "invalid"
        raise ValueError("Unknown RoundingMode enum entry.")



@register_attribute_builder("RoundingModeAttr", allow_existing=True)
def _roundingmodeattr(x, context):
    return _ods_ir.IntegerAttr.get(_ods_ir.IntegerType.get_signless(64, context=context), int(x))

class TailCallKind(IntEnum):
    """Tail Call Kind"""

    None_ = 0
    NoTail = 3
    MustTail = 2
    Tail = 1

    def __str__(self):
        if self is TailCallKind.None_:
            return "none"
        if self is TailCallKind.NoTail:
            return "notail"
        if self is TailCallKind.MustTail:
            return "musttail"
        if self is TailCallKind.Tail:
            return "tail"
        raise ValueError("Unknown TailCallKind enum entry.")



@register_attribute_builder("TailCallKindEnum", allow_existing=True)
def _tailcallkindenum(x, context):
    return _ods_ir.IntegerAttr.get(_ods_ir.IntegerType.get_signless(64, context=context), int(x))

class UWTableKind(IntEnum):
    """LLVM Unwind Behavior"""

    None_ = 0
    Sync = 1
    Async = 2

    def __str__(self):
        if self is UWTableKind.None_:
            return "none"
        if self is UWTableKind.Sync:
            return "sync"
        if self is UWTableKind.Async:
            return "async"
        raise ValueError("Unknown UWTableKind enum entry.")



@register_attribute_builder("UWTableKindEnum", allow_existing=True)
def _uwtablekindenum(x, context):
    return _ods_ir.IntegerAttr.get(_ods_ir.IntegerType.get_signless(64, context=context), int(x))

class UnnamedAddr(IntEnum):
    """LLVM GlobalValue UnnamedAddr"""

    None_ = 0
    Local = 1
    Global = 2

    def __str__(self):
        if self is UnnamedAddr.None_:
            return ""
        if self is UnnamedAddr.Local:
            return "local_unnamed_addr"
        if self is UnnamedAddr.Global:
            return "unnamed_addr"
        raise ValueError("Unknown UnnamedAddr enum entry.")



@register_attribute_builder("UnnamedAddr", allow_existing=True)
def _unnamedaddr(x, context):
    return _ods_ir.IntegerAttr.get(_ods_ir.IntegerType.get_signless(64, context=context), int(x))

class Visibility(IntEnum):
    """LLVM GlobalValue Visibility"""

    Default = 0
    Hidden = 1
    Protected = 2

    def __str__(self):
        if self is Visibility.Default:
            return ""
        if self is Visibility.Hidden:
            return "hidden"
        if self is Visibility.Protected:
            return "protected"
        raise ValueError("Unknown Visibility enum entry.")



@register_attribute_builder("Visibility", allow_existing=True)
def _visibility(x, context):
    return _ods_ir.IntegerAttr.get(_ods_ir.IntegerType.get_signless(64, context=context), int(x))

@register_attribute_builder("llvm.LLVM_FastmathFlagsAttr")
def _llvm_fastmathflagsattr(x, context):
    return _ods_ir.Attribute.parse(f'#llvm.fastmath<{str(x)}>', context=context)

@register_attribute_builder("llvm.LLVM_IntegerOverflowFlagsAttr")
def _llvm_integeroverflowflagsattr(x, context):
    return _ods_ir.Attribute.parse(f'#llvm.overflow<{str(x)}>', context=context)

